import asyncio
import logging
from datetime import datetime, timezone
from typing import Any, Tuple, List, Optional

from dev_health_ops.models.git import (
    GitBlame,
    GitCommit,
    GitCommitStat,
    GitFile,
    GitPullRequest,
    GitPullRequestReview,
    CiPipelineRun,
    Deployment,
    Incident,
    Repo,
)
from dev_health_ops.utils import (
    AGGREGATE_STATS_MARKER,
    BATCH_SIZE,
    CONNECTORS_AVAILABLE,
    is_skippable,
)

if CONNECTORS_AVAILABLE:
    from dev_health_ops.connectors import (
        BatchResult,
        GitHubConnector,
        ConnectorException,
    )
    from dev_health_ops.connectors.models import Repository
    from dev_health_ops.connectors.utils import RateLimitConfig, RateLimitGate
    from github import RateLimitExceededException
else:
    BatchResult = None  # type: ignore
    GitHubConnector = None  # type: ignore
    ConnectorException = Exception
    Repository = None  # type: ignore
    RateLimitConfig = None  # type: ignore
    RateLimitGate = None  # type: ignore
    RateLimitExceededException = Exception


# --- GitHub Sync Helpers ---


def _fetch_github_repo_info_sync(connector, owner, repo_name):
    """Sync helper to fetch GitHub repository info."""
    gh_repo = connector.github.get_repo(f"{owner}/{repo_name}")
    _ = gh_repo.id
    return gh_repo


def _fetch_github_commits_sync(
    gh_repo,
    max_commits: Optional[int],
    repo_id,
    since: Optional[datetime] = None,
):
    """Sync helper to fetch and parse GitHub commits."""
    raw_commits = []
    if since is not None:
        commits_iter = gh_repo.get_commits(since=since)
    else:
        commits_iter = gh_repo.get_commits()

    for commit in commits_iter:
        raw_commits.append(commit)
        if max_commits is not None and len(raw_commits) >= max_commits:
            break

    commit_objects = []
    for commit in raw_commits:
        if since is not None:
            commit_when = None
            if getattr(commit, "commit", None) and getattr(
                commit.commit, "committer", None
            ):
                commit_when = getattr(commit.commit.committer, "date", None)
            if (
                commit_when is None
                and getattr(commit, "commit", None)
                and getattr(commit.commit, "author", None)
            ):
                commit_when = getattr(commit.commit.author, "date", None)

            if (
                isinstance(commit_when, datetime)
                and commit_when.astimezone(timezone.utc) < since
            ):
                continue

        # Prefer GitHub user `login` when available; do not store emails.
        author_login = getattr(commit, "author", None)
        committer_login = getattr(commit, "committer", None)

        author_name = getattr(author_login, "login", None) or (
            commit.commit.author.name if commit.commit.author else "Unknown"
        )
        committer_name = getattr(committer_login, "login", None) or (
            commit.commit.committer.name if commit.commit.committer else "Unknown"
        )

        # Safely obtain emails: prefer commit metadata (no extra API calls),
        # fallback to user.email but guard against API-triggered exceptions (e.g., 404).
        def _safe_user_email(user):
            try:
                return getattr(user, "email", None)
            except Exception:
                return None

        author_email = None
        if getattr(commit, "commit", None) and getattr(commit.commit, "author", None):
            author_email = getattr(commit.commit.author, "email", None)
        if not author_email:
            author_email = _safe_user_email(author_login)

        committer_email = None
        if getattr(commit, "commit", None) and getattr(
            commit.commit, "committer", None
        ):
            committer_email = getattr(commit.commit.committer, "email", None)
        if not committer_email:
            committer_email = _safe_user_email(committer_login)

        git_commit = GitCommit(
            repo_id=repo_id,
            hash=commit.sha,
            message=commit.commit.message,
            author_name=author_name,
            author_email=author_email,
            author_when=(
                commit.commit.author.date
                if commit.commit.author
                else datetime.now(timezone.utc)
            ),
            committer_name=committer_name,
            committer_email=committer_email,
            committer_when=(
                commit.commit.committer.date
                if commit.commit.committer
                else datetime.now(timezone.utc)
            ),
            parents=len(commit.parents),
        )
        commit_objects.append(git_commit)
    return raw_commits, commit_objects


def _fetch_github_commit_stats_sync(
    raw_commits,
    repo_id,
    max_stats,
    since: Optional[datetime] = None,
):
    """Sync helper to fetch detailed commit stats (files)."""
    stats_objects = []
    for commit in raw_commits[:max_stats]:
        if since is not None:
            commit_when = None
            if getattr(commit, "commit", None) and getattr(
                commit.commit, "committer", None
            ):
                commit_when = getattr(commit.commit.committer, "date", None)
            if (
                commit_when is None
                and getattr(commit, "commit", None)
                and getattr(commit.commit, "author", None)
            ):
                commit_when = getattr(commit.commit.author, "date", None)

            if (
                isinstance(commit_when, datetime)
                and commit_when.astimezone(timezone.utc) < since
            ):
                continue

        try:
            files = commit.files
            if files is None:
                continue

            for file in files:
                stat = GitCommitStat(
                    repo_id=repo_id,
                    commit_hash=commit.sha,
                    file_path=file.filename,
                    additions=file.additions,
                    deletions=file.deletions,
                    old_file_mode="unknown",
                    new_file_mode="unknown",
                )
                stats_objects.append(stat)
        except Exception as e:
            logging.warning(
                "Failed to get stats for commit %s: %s",
                commit.sha,
                e,
            )
    return stats_objects


def _fetch_github_prs_sync(connector, owner, repo_name, repo_id, max_prs):
    """Sync helper to fetch Pull Requests."""
    prs = connector.get_pull_requests(
        owner,
        repo_name,
        state="all",
        max_prs=max_prs,
    )
    pr_objects = []
    for pr in prs:
        created_at = (
            pr.created_at or pr.merged_at or pr.closed_at or datetime.now(timezone.utc)
        )
        git_pr = GitPullRequest(
            repo_id=repo_id,
            number=pr.number,
            title=pr.title,
            body=pr.body,
            state=pr.state,
            author_name=pr.author.username if pr.author else "Unknown",
            author_email=pr.author.email if pr.author else None,
            created_at=created_at,
            merged_at=pr.merged_at,
            closed_at=pr.closed_at,
            head_branch=pr.head_branch,
            base_branch=pr.base_branch,
        )
        pr_objects.append(git_pr)
    return pr_objects


def _fetch_github_workflow_runs_sync(gh_repo, repo_id, max_runs, since):
    runs = []
    if not hasattr(gh_repo, "get_workflow_runs"):
        if not hasattr(gh_repo, "get_workflows"):
            return runs

    def _coerce_datetime(value):
        if isinstance(value, datetime):
            return value
        if isinstance(value, str):
            try:
                return datetime.fromisoformat(value.replace("Z", "+00:00"))
            except Exception:
                return None
        return None

    try:
        if hasattr(gh_repo, "get_workflow_runs"):
            raw_runs = []
            for run in gh_repo.get_workflow_runs():
                raw_runs.append(run)
                if len(raw_runs) >= max_runs:
                    break
        else:
            raw_runs = []
            for workflow in gh_repo.get_workflows():
                for run in workflow.get_runs():
                    raw_runs.append(run)
                    if len(raw_runs) >= max_runs:
                        break
                if len(raw_runs) >= max_runs:
                    break
    except Exception as exc:
        logging.debug("Failed to fetch workflow runs: %s", exc)
        return runs

    for run in raw_runs:
        queued_at = _coerce_datetime(getattr(run, "created_at", None))
        started_at = _coerce_datetime(getattr(run, "run_started_at", None)) or queued_at
        if started_at is None:
            continue
        if since is not None and started_at.astimezone(timezone.utc) < since:
            continue
        finished_at = _coerce_datetime(getattr(run, "updated_at", None))
        runs.append(
            CiPipelineRun(
                repo_id=repo_id,
                run_id=str(getattr(run, "id", "")),
                status=getattr(run, "conclusion", None) or getattr(run, "status", None),
                queued_at=queued_at,
                started_at=started_at,
                finished_at=finished_at,
            )
        )
    return runs


def _fetch_github_deployments_sync(gh_repo, repo_id, max_deployments, since):
    deployments = []
    if not hasattr(gh_repo, "get_deployments"):
        return deployments
    try:
        raw_deployments = list(gh_repo.get_deployments()[:max_deployments])
    except Exception as exc:
        logging.debug("Failed to fetch deployments: %s", exc)
        return deployments

    for dep in raw_deployments:
        created_at = getattr(dep, "created_at", None)
        if not isinstance(created_at, datetime):
            continue
        if since is not None and created_at.astimezone(timezone.utc) < since:
            continue
        deployments.append(
            Deployment(
                repo_id=repo_id,
                deployment_id=str(getattr(dep, "id", "")),
                status=getattr(dep, "state", None),
                environment=getattr(dep, "environment", None),
                started_at=created_at,
                finished_at=None,
                deployed_at=created_at,
                merged_at=None,
                pull_request_number=None,
            )
        )
    return deployments


def _fetch_github_incidents_sync(gh_repo, repo_id, max_issues, since):
    incidents = []
    if not hasattr(gh_repo, "get_issues"):
        return incidents
    try:
        raw_issues = list(
            gh_repo.get_issues(state="all", labels=["incident"])[:max_issues]
        )
    except Exception as exc:
        logging.debug("Failed to fetch incident issues: %s", exc)
        return incidents

    for issue in raw_issues:
        created_at = getattr(issue, "created_at", None)
        if not isinstance(created_at, datetime):
            continue
        if since is not None and created_at.astimezone(timezone.utc) < since:
            continue
        incidents.append(
            Incident(
                repo_id=repo_id,
                incident_id=str(getattr(issue, "id", "")),
                status=getattr(issue, "state", None),
                started_at=created_at,
                resolved_at=getattr(issue, "closed_at", None),
            )
        )
    return incidents


def _sync_github_prs_to_store(
    connector,
    owner: str,
    repo_name: str,
    repo_id,
    store,
    loop: asyncio.AbstractEventLoop,
    batch_size: int,
    state: str = "all",
    gate: Optional[RateLimitGate] = None,
    since: Optional[datetime] = None,
) -> int:
    """Fetch all PRs for a repo and insert them in batches.

    Runs in a worker thread; uses run_coroutine_threadsafe to write batches.
    """
    logging.info(
        "Fetching PRs for %s/%s...",
        owner,
        repo_name,
    )
    gh_repo = connector.github.get_repo(f"{owner}/{repo_name}")
    batch: List[GitPullRequest] = []
    total = 0

    if gate is None:
        gate = RateLimitGate(RateLimitConfig(initial_backoff_seconds=1.0))

    # per_page is set at Github client level during connector initialization
    # (some test doubles or older PyGithub versions may not accept sort/direction).
    sorted_by_updated = True
    try:
        pr_iter = iter(gh_repo.get_pulls(state=state, sort="updated", direction="desc"))
    except TypeError:
        sorted_by_updated = False
        pr_iter = iter(gh_repo.get_pulls(state=state))
    while True:
        try:
            gate.wait_sync()
            gh_pr = next(pr_iter)
            gate.reset()
        except StopIteration:
            break
        except RateLimitExceededException as e:
            retry_after = None
            if hasattr(connector, "_rate_limit_reset_delay_seconds"):
                try:
                    retry_after = connector._rate_limit_reset_delay_seconds()
                except Exception:
                    retry_after = None
            if retry_after is None:
                headers = getattr(e, "headers", None)
                if isinstance(headers, dict):
                    headers_ci = {str(k).lower(): v for k, v in headers.items()}
                    ra = headers_ci.get("retry-after")
                    if ra is not None:
                        try:
                            retry_after = float(ra)
                        except ValueError:
                            retry_after = None
            applied = gate.penalize(retry_after)
            logging.info(
                "GitHub rate limited fetching PRs; backoff %.1fs (%s)",
                applied,
                e,
            )
            continue
        except Exception as e:
            headers = getattr(e, "headers", None)
            retry_after = None
            if isinstance(headers, dict):
                # PyGithub header casing varies; treat keys case-insensitively.
                headers_ci = {str(k).lower(): v for k, v in headers.items()}
                ra = headers_ci.get("retry-after")
                if ra:
                    try:
                        retry_after = float(ra)
                    except ValueError:
                        retry_after = None
                if retry_after is None:
                    reset = headers_ci.get("x-ratelimit-reset")
                    if reset:
                        try:
                            import time

                            retry_after = max(0.0, float(reset) - time.time())
                        except ValueError:
                            retry_after = None

            if retry_after is not None:
                applied = gate.penalize(retry_after)
                logging.info(
                    "GitHub rate limited fetching PRs; backoff %.1fs (%s)",
                    applied,
                    e,
                )
                continue

            raise
        if since is not None and sorted_by_updated:
            updated_at = getattr(gh_pr, "updated_at", None)
            if (
                isinstance(updated_at, datetime)
                and updated_at.astimezone(timezone.utc) < since
            ):
                break

        author_name = "Unknown"
        author_email = None
        if getattr(gh_pr, "user", None):
            author_name = getattr(gh_pr.user, "login", None) or author_name

        created_at = (
            getattr(gh_pr, "created_at", None)
            or getattr(gh_pr, "merged_at", None)
            or getattr(gh_pr, "closed_at", None)
            or datetime.now(timezone.utc)
        )

        merged_at = getattr(gh_pr, "merged_at", None)
        closed_at = getattr(gh_pr, "closed_at", None)

        # Lazy load full PR for stats (additions/deletions)
        additions = getattr(gh_pr, "additions", 0)
        deletions = getattr(gh_pr, "deletions", 0)
        changed_files = getattr(gh_pr, "changed_files", 0)

        # Fetch and store reviews for this PR
        first_review_at = None
        reviews_count = 0
        changes_requested_count = 0
        try:
            reviews = connector.get_pull_request_reviews(owner, repo_name, gh_pr.number)
            reviews_count = len(reviews)
            if reviews:
                review_objects = []
                for r in reviews:
                    review_at = r.submitted_at or created_at
                    if first_review_at is None or review_at < first_review_at:
                        first_review_at = review_at
                    if r.state == "CHANGES_REQUESTED":
                        changes_requested_count += 1

                    review_objects.append(
                        GitPullRequestReview(
                            repo_id=repo_id,
                            number=gh_pr.number,
                            review_id=r.id,
                            reviewer=r.reviewer,
                            state=r.state,
                            submitted_at=review_at,
                        )
                    )
                asyncio.run_coroutine_threadsafe(
                    store.insert_git_pull_request_reviews(review_objects),
                    loop,
                ).result()
        except Exception as e:
            logging.debug(f"Failed to fetch reviews for PR #{gh_pr.number}: {e}")

        # Fetch first comment for "Pickup Time"
        first_comment_at = None
        comments_count = 0
        try:
            # Issue comments include top-level PR comments
            issue_comments = gh_pr.get_issue_comments()
            # This is a paginated list, we only need the first one's date but we want total count
            # To avoid fetching all comments if there are many, we can just get the first page for first_comment_at
            # and maybe the count is available on the issue object.
            # However, for simplicity and since we are already Doing one call per PR, let's just get the first one.
            comments_count = gh_pr.comments  # issue comments count
            if comments_count > 0:
                for c in issue_comments:
                    if first_comment_at is None or c.created_at < first_comment_at:
                        first_comment_at = c.created_at
                    # if we only want the first one, we can break if they are sorted,
                    # but they might not be.
        except Exception as e:
            logging.debug(f"Failed to fetch comments for PR #{gh_pr.number}: {e}")

        batch.append(
            GitPullRequest(
                repo_id=repo_id,
                number=int(getattr(gh_pr, "number", 0) or 0),
                title=getattr(gh_pr, "title", None),
                body=getattr(gh_pr, "body", None),
                state=getattr(gh_pr, "state", None),
                author_name=author_name,
                author_email=author_email,
                created_at=created_at,
                merged_at=merged_at,
                closed_at=closed_at,
                head_branch=getattr(getattr(gh_pr, "head", None), "ref", None),
                base_branch=getattr(getattr(gh_pr, "base", None), "ref", None),
                additions=additions,
                deletions=deletions,
                changed_files=changed_files,
                first_review_at=first_review_at,
                first_comment_at=first_comment_at,
                changes_requested_count=changes_requested_count,
                reviews_count=reviews_count,
                comments_count=comments_count,
            )
        )
        total += 1

        if len(batch) >= batch_size:
            asyncio.run_coroutine_threadsafe(
                store.insert_git_pull_requests(batch),
                loop,
            ).result()
            logging.debug(
                "Stored batch of %d PRs for %s/%s (total: %d)",
                len(batch),
                owner,
                repo_name,
                total,
            )
            batch.clear()

    if batch:
        asyncio.run_coroutine_threadsafe(
            store.insert_git_pull_requests(batch),
            loop,
        ).result()

    logging.info(
        "Fetched %d PRs for %s/%s",
        total,
        owner,
        repo_name,
    )

    return total


def _fetch_github_blame_sync(gh_repo, repo_id, limit=50):
    """Sync helper to fetch (simulated) blame by listing files."""
    files_to_process = []
    try:
        contents = gh_repo.get_contents("", ref=gh_repo.default_branch)
        while contents:
            file_content = contents.pop(0)
            if file_content.type == "dir":
                contents.extend(
                    gh_repo.get_contents(
                        file_content.path,
                        ref=gh_repo.default_branch,
                    )
                )
            else:
                if not is_skippable(file_content.path):
                    files_to_process.append(file_content.path)

            if len(files_to_process) >= limit:
                break
    except Exception as e:
        logging.error(f"Error listing files: {e}")
    return []


def _split_full_name(full_name: str) -> Tuple[str, str]:
    parts = (full_name or "").split("/", 1)
    if len(parts) != 2:
        raise ValueError(f"Invalid repo/project full name: {full_name}")
    return parts[0], parts[1]


async def _backfill_github_missing_data(
    store: Any,
    connector: GitHubConnector,
    db_repo: Repo,
    repo_full_name: str,
    default_branch: str,
    max_commits: Optional[int],
    blame_only: bool = False,
) -> None:
    # Logic matches the CLI sync orchestration.
    logging.info(
        "Backfilling data for %s...",
        repo_full_name,
    )
    owner, repo_name = _split_full_name(repo_full_name)

    if not (
        hasattr(store, "has_any_git_files")
        and hasattr(store, "has_any_git_blame")
        and hasattr(store, "has_any_git_commit_stats")
    ):
        return

    needs_files = not await store.has_any_git_files(db_repo.id)
    needs_commit_stats = (
        False if blame_only else not await store.has_any_git_commit_stats(db_repo.id)
    )
    needs_blame = not await store.has_any_git_blame(db_repo.id)

    if not (needs_files or needs_commit_stats or needs_blame):
        return

    gh_repo = connector.github.get_repo(f"{owner}/{repo_name}")

    file_paths: List[str] = []
    if needs_files or needs_blame:
        try:
            branch = gh_repo.get_branch(default_branch)
            tree = gh_repo.get_git_tree(branch.commit.sha, recursive=True)
            for entry in getattr(tree, "tree", []) or []:
                if getattr(entry, "type", None) != "blob":
                    continue
                path = getattr(entry, "path", None)
                if not path:
                    continue
                file_paths.append(path)

            if needs_files and file_paths:
                batch: List[GitFile] = []
                for path in file_paths:
                    batch.append(
                        GitFile(
                            repo_id=db_repo.id,
                            path=path,
                            executable=False,
                            contents=None,
                        )
                    )
                    if len(batch) >= BATCH_SIZE:
                        await store.insert_git_file_data(batch)
                        batch.clear()
                if batch:
                    await store.insert_git_file_data(batch)
                logging.info(
                    "Backfilled %d files for %s",
                    len(file_paths),
                    repo_full_name,
                )
        except Exception as e:
            logging.warning(
                f"Failed to backfill GitHub files for {repo_full_name}: {e}"
            )

    if needs_commit_stats:
        try:
            logging.info(
                "Backfilling commit stats for %s...",
                repo_full_name,
            )
            commits_iter = gh_repo.get_commits()
            commit_stats_batch: List[GitCommitStat] = []
            commit_count = 0
            for commit in commits_iter:
                if max_commits and commit_count >= max_commits:
                    break
                commit_count += 1
                try:
                    detailed = gh_repo.get_commit(commit.sha)
                    for file in getattr(detailed, "files", []) or []:
                        commit_stats_batch.append(
                            GitCommitStat(
                                repo_id=db_repo.id,
                                commit_hash=commit.sha,
                                file_path=getattr(
                                    file, "filename", AGGREGATE_STATS_MARKER
                                ),
                                additions=getattr(file, "additions", 0),
                                deletions=getattr(file, "deletions", 0),
                                old_file_mode="unknown",
                                new_file_mode="unknown",
                            )
                        )
                        if len(commit_stats_batch) >= BATCH_SIZE:
                            await store.insert_git_commit_stats(commit_stats_batch)
                            logging.debug(
                                "Stored batch of %d commit stats for %s",
                                len(commit_stats_batch),
                                repo_full_name,
                            )
                            commit_stats_batch.clear()
                except Exception as e:
                    logging.debug(
                        "Failed commit stat fetch for %s@%s: %s",
                        repo_full_name,
                        commit.sha,
                        e,
                    )

            if commit_stats_batch:
                await store.insert_git_commit_stats(commit_stats_batch)
            logging.info(
                "Backfilled commit stats for %d commits in %s",
                commit_count,
                repo_full_name,
            )
        except Exception as e:
            logging.warning(
                "Failed to backfill GitHub commit stats for %s: %s",
                repo_full_name,
                e,
            )

    if needs_blame and file_paths:
        try:
            logging.info(
                "Backfilling blame for %d files in %s...",
                len(file_paths),
                repo_full_name,
            )
            blame_batch: List[GitBlame] = []
            processed_files = 0
            for path in file_paths:
                try:
                    blame = connector.get_file_blame(
                        owner=owner,
                        repo=repo_name,
                        path=path,
                        ref=default_branch,
                    )
                    processed_files += 1
                except Exception as e:
                    logging.debug(
                        f"Failed blame fetch for {repo_full_name}:{path}: {e}"
                    )
                    continue

                for rng in blame.ranges:
                    for line_no in range(
                        rng.starting_line,
                        rng.ending_line + 1,
                    ):
                        blame_batch.append(
                            GitBlame(
                                repo_id=db_repo.id,
                                path=path,
                                line_no=line_no,
                                author_email=rng.author_email,
                                author_name=rng.author,
                                author_when=None,
                                commit_hash=rng.commit_sha,
                                line=None,
                            )
                        )
                        if len(blame_batch) >= BATCH_SIZE:
                            await store.insert_blame_data(blame_batch)
                            logging.debug(
                                "Stored batch of %d blame entries for %s",
                                len(blame_batch),
                                repo_full_name,
                            )
                            blame_batch.clear()

            if blame_batch:
                await store.insert_blame_data(blame_batch)
            logging.info(
                "Backfilled blame for %d files in %s",
                processed_files,
                repo_full_name,
            )
        except Exception as e:
            logging.warning(
                f"Failed to backfill GitHub blame for {repo_full_name}: {e}"
            )


async def process_github_repo(
    store: Any,
    owner: str,
    repo_name: str,
    token: str,
    fetch_blame: bool = False,
    blame_only: bool = False,
    max_commits: Optional[int] = None,
    sync_git: bool = True,
    sync_prs: bool = True,
    sync_cicd: bool = True,
    sync_deployments: bool = True,
    sync_incidents: bool = True,
    since: Optional[datetime] = None,
) -> None:
    """
    Process a GitHub repository using the GitHub connector.
    """
    if not CONNECTORS_AVAILABLE:
        raise RuntimeError("Connectors unavailable. Install required dependencies.")

    logging.info(f"Processing GitHub repository: {owner}/{repo_name}")
    loop = asyncio.get_running_loop()

    connector = GitHubConnector(token=token)
    try:
        # 1. Fetch Repo Info
        logging.info("Fetching repository information...")
        gh_repo = await loop.run_in_executor(
            None, _fetch_github_repo_info_sync, connector, owner, repo_name
        )

        # Create/Insert Repo
        repo_info = Repository(
            id=gh_repo.id,
            name=gh_repo.name,
            full_name=gh_repo.full_name,
            default_branch=gh_repo.default_branch,
            description=gh_repo.description,
            url=gh_repo.html_url,
            created_at=gh_repo.created_at,
            updated_at=gh_repo.updated_at,
            language=gh_repo.language,
            stars=gh_repo.stargazers_count,
            forks=gh_repo.forks_count,
        )

        db_repo = Repo(
            repo_path=None,
            repo=repo_info.full_name,
            settings={
                "source": "github",
                "repo_id": repo_info.id,
                "url": repo_info.url,
                "default_branch": repo_info.default_branch,
            },
            tags=[
                "github",
                repo_info.language,
            ]
            if repo_info.language
            else ["github"],
        )

        await store.insert_repo(db_repo)
        logging.info(f"Repository stored: {db_repo.repo} ({db_repo.id})")

        if blame_only:
            await _backfill_github_missing_data(
                store=store,
                connector=connector,
                db_repo=db_repo,
                repo_full_name=db_repo.repo,
                default_branch=repo_info.default_branch,
                max_commits=max_commits,
                blame_only=True,
            )
            logging.info(
                "Completed blame-only sync for GitHub repository: %s/%s",
                owner,
                repo_name,
            )
            return

        if sync_git:
            # 2. Fetch Commits
            if max_commits is None:
                logging.info("Fetching all commits from GitHub...")
            else:
                logging.info(f"Fetching up to {max_commits} commits from GitHub...")
            raw_commits, commit_objects = await loop.run_in_executor(
                None,
                _fetch_github_commits_sync,
                gh_repo,
                max_commits,
                db_repo.id,
                since,
            )

            if commit_objects:
                await store.insert_git_commit_data(commit_objects)
                logging.info(f"Stored {len(commit_objects)} commits from GitHub")

            # 3. Fetch Stats
            logging.info("Fetching commit stats from GitHub...")
            stats_limit = 50 if max_commits is None else min(max_commits, 50)
            stats_objects = await loop.run_in_executor(
                None,
                _fetch_github_commit_stats_sync,
                raw_commits,
                db_repo.id,
                stats_limit,
                since,
            )

            if stats_objects:
                await store.insert_git_commit_stats(stats_objects)
                logging.info(
                    "Stored %d commit stats from GitHub",
                    len(stats_objects),
                )

        if sync_prs:
            # 4. Fetch PRs
            logging.info("Fetching pull requests from GitHub...")
            pr_total = await loop.run_in_executor(
                None,
                _sync_github_prs_to_store,
                connector,
                owner,
                repo_name,
                db_repo.id,
                store,
                loop,
                BATCH_SIZE,
                "all",
                None,
                since,
            )
            logging.info(f"Stored {pr_total} pull requests from GitHub")

        if sync_cicd:
            logging.info("Fetching CI/CD workflow runs from GitHub...")
            pipeline_runs = await loop.run_in_executor(
                None,
                _fetch_github_workflow_runs_sync,
                gh_repo,
                db_repo.id,
                BATCH_SIZE,
                since,
            )
            if pipeline_runs:
                await store.insert_ci_pipeline_runs(pipeline_runs)
                logging.info("Stored %d workflow runs", len(pipeline_runs))

        if sync_deployments:
            logging.info("Fetching deployments from GitHub...")
            deployments = await loop.run_in_executor(
                None,
                _fetch_github_deployments_sync,
                gh_repo,
                db_repo.id,
                BATCH_SIZE,
                since,
            )
            if deployments:
                await store.insert_deployments(deployments)
                logging.info("Stored %d deployments", len(deployments))

        if sync_incidents:
            logging.info("Fetching incident issues from GitHub...")
            incidents = await loop.run_in_executor(
                None,
                _fetch_github_incidents_sync,
                gh_repo,
                db_repo.id,
                BATCH_SIZE,
                since,
            )
            if incidents:
                await store.insert_incidents(incidents)
                logging.info("Stored %d incidents", len(incidents))

        # 5. Fetch Blame (Optional & Stubbed)
        if fetch_blame:
            logging.info("Fetching blame data (file list) from GitHub...")
            await loop.run_in_executor(
                None, _fetch_github_blame_sync, gh_repo, db_repo.id
            )

        logging.info(
            "Successfully processed GitHub repository: %s/%s",
            owner,
            repo_name,
        )

    except ConnectorException as e:
        logging.error(f"Connector error: {e}")
        raise
    except Exception as e:
        logging.error(f"Error processing GitHub repository: {e}")
        raise
    finally:
        connector.close()


async def process_github_repos_batch(
    store: Any,
    token: str,
    org_name: str = None,
    user_name: str = None,
    pattern: str = None,
    batch_size: int = 10,
    max_concurrent: int = 4,
    rate_limit_delay: float = 1.0,
    max_commits_per_repo: int = None,
    max_repos: int = None,
    use_async: bool = False,
    sync_git: bool = True,
    sync_prs: bool = True,
    sync_cicd: bool = True,
    sync_deployments: bool = True,
    sync_incidents: bool = True,
    blame_only: bool = False,
    backfill_missing: bool = True,
    since: Optional[datetime] = None,
) -> None:
    """
    Process multiple GitHub repositories using batch processing with
    pattern matching.
    """
    if not CONNECTORS_AVAILABLE:
        raise RuntimeError("Connectors unavailable. Install required dependencies.")

    logging.info("=== GitHub Batch Repository Processing ===")
    connector = GitHubConnector(token=token)
    loop = asyncio.get_running_loop()

    pr_gate = None
    pr_semaphore = None
    if sync_prs:
        pr_gate = RateLimitGate(
            RateLimitConfig(initial_backoff_seconds=max(1.0, rate_limit_delay))
        )
        pr_semaphore = asyncio.Semaphore(max(1, max_concurrent))

    # Track results for summary and incremental storage
    all_results: List[BatchResult] = []
    stored_count = 0

    results_queue: Optional[asyncio.Queue] = None
    _queue_sentinel = object()

    async def store_result(result: BatchResult) -> None:
        """Store a single result in the database (upsert)."""
        nonlocal stored_count
        if not result.success:
            return

        repo_info = result.repository
        db_repo = Repo(
            repo_path=None,  # Not a local repo
            repo=repo_info.full_name,
            settings={
                "source": "github",
                "repo_id": repo_info.id,
                "url": repo_info.url,
                "default_branch": repo_info.default_branch,
                "batch_processed": True,
            },
            tags=[
                "github",
                repo_info.language,
            ]
            if repo_info.language
            else ["github"],
        )

        await store.insert_repo(db_repo)
        stored_count += 1
        logging.debug(f"Stored repository ({stored_count}): {db_repo.repo}")

        if blame_only:
            try:
                await _backfill_github_missing_data(
                    store=store,
                    connector=connector,
                    db_repo=db_repo,
                    repo_full_name=repo_info.full_name,
                    default_branch=repo_info.default_branch,
                    max_commits=max_commits_per_repo,
                    blame_only=True,
                )
            except Exception as e:
                logging.debug(
                    "Blame-only backfill failed for GitHub repo %s: %s",
                    repo_info.full_name,
                    e,
                )
            return

        gh_repo = None
        if sync_git:
            # Fetch commits and stats to populate git_commits/git_commit_stats.
            if max_commits_per_repo is None and since is None:
                commit_limit = 100
            else:
                commit_limit = max_commits_per_repo
            try:
                if gh_repo is None:
                    gh_repo = connector.github.get_repo(repo_info.full_name)
                raw_commits, commit_objects = await loop.run_in_executor(
                    None,
                    _fetch_github_commits_sync,
                    gh_repo,
                    commit_limit,
                    db_repo.id,
                    since,
                )
                if commit_objects:
                    await store.insert_git_commit_data(commit_objects)

                stats_objects = await loop.run_in_executor(
                    None,
                    _fetch_github_commit_stats_sync,
                    raw_commits,
                    db_repo.id,
                    50 if commit_limit is None else min(commit_limit, 50),
                    since,
                )
                if stats_objects:
                    await store.insert_git_commit_stats(stats_objects)
            except Exception as e:
                logging.warning(
                    "Failed to fetch commits for GitHub repo %s: %s",
                    repo_info.full_name,
                    e,
                    exc_info=True,
                )

        if sync_prs:
            # Fetch ALL PRs for batch-processed repos, storing in batches.
            try:
                owner, repo_name = _split_full_name(repo_info.full_name)
                async with pr_semaphore:
                    await loop.run_in_executor(
                        None,
                        _sync_github_prs_to_store,
                        connector,
                        owner,
                        repo_name,
                        db_repo.id,
                        store,
                        loop,
                        BATCH_SIZE,
                        "all",
                        pr_gate,
                        since,
                    )
            except Exception as e:
                logging.error(
                    "Failed to fetch/store PRs for GitHub repo %s: %s",
                    repo_info.full_name,
                    e,
                )
                raise

        if sync_cicd:
            try:
                if gh_repo is None:
                    gh_repo = connector.github.get_repo(repo_info.full_name)
                pipeline_runs = await loop.run_in_executor(
                    None,
                    _fetch_github_workflow_runs_sync,
                    gh_repo,
                    db_repo.id,
                    BATCH_SIZE,
                    since,
                )
                if pipeline_runs:
                    await store.insert_ci_pipeline_runs(pipeline_runs)
            except Exception as e:
                logging.warning(
                    "Failed to fetch CI/CD runs for GitHub repo %s: %s",
                    repo_info.full_name,
                    e,
                )

        if sync_deployments:
            try:
                if gh_repo is None:
                    gh_repo = connector.github.get_repo(repo_info.full_name)
                deployments = await loop.run_in_executor(
                    None,
                    _fetch_github_deployments_sync,
                    gh_repo,
                    db_repo.id,
                    BATCH_SIZE,
                    since,
                )
                if deployments:
                    await store.insert_deployments(deployments)
            except Exception as e:
                logging.warning(
                    "Failed to fetch deployments for GitHub repo %s: %s",
                    repo_info.full_name,
                    e,
                )

        if sync_incidents:
            try:
                if gh_repo is None:
                    gh_repo = connector.github.get_repo(repo_info.full_name)
                incidents = await loop.run_in_executor(
                    None,
                    _fetch_github_incidents_sync,
                    gh_repo,
                    db_repo.id,
                    BATCH_SIZE,
                    since,
                )
                if incidents:
                    await store.insert_incidents(incidents)
            except Exception as e:
                logging.warning(
                    "Failed to fetch incidents for GitHub repo %s: %s",
                    repo_info.full_name,
                    e,
                )

        if result.stats and sync_git:
            stat = GitCommitStat(
                repo_id=db_repo.id,
                commit_hash=AGGREGATE_STATS_MARKER,
                file_path=AGGREGATE_STATS_MARKER,
                additions=result.stats.additions,
                deletions=result.stats.deletions,
                old_file_mode="unknown",
                new_file_mode="unknown",
            )
            await store.insert_git_commit_stats([stat])

        if backfill_missing:
            try:
                await _backfill_github_missing_data(
                    store=store,
                    connector=connector,
                    db_repo=db_repo,
                    repo_full_name=repo_info.full_name,
                    default_branch=repo_info.default_branch,
                    max_commits=max_commits_per_repo,
                )
            except Exception as e:
                logging.debug(
                    "Backfill failed for GitHub repo %s: %s",
                    repo_info.full_name,
                    e,
                )

    def on_repo_complete(result: BatchResult) -> None:
        all_results.append(result)
        if result.success:
            stats_info = ""
            if result.stats:
                stats_info = f" ({result.stats.total_commits} commits)"
            logging.info(
                "  \u2713 Processed: %s%s",
                result.repository.full_name,
                stats_info,
            )
        else:
            logging.warning(
                f"  ✗ Failed: {result.repository.full_name}: {result.error}"
            )

        if results_queue is not None:
            try:
                running_loop = asyncio.get_running_loop()
            except RuntimeError:
                running_loop = None

            def _enqueue() -> None:
                assert results_queue is not None
                try:
                    results_queue.put_nowait(result)
                except asyncio.QueueFull:
                    asyncio.create_task(results_queue.put(result))

            if running_loop is loop:
                _enqueue()
            else:
                loop.call_soon_threadsafe(_enqueue)

    try:
        if sync_git:
            results_queue = asyncio.Queue(maxsize=max(1, max_concurrent * 2))

            async def _consume_results() -> None:
                assert results_queue is not None
                while True:
                    item = await results_queue.get()
                    try:
                        if item is _queue_sentinel:
                            return
                        await store_result(item)
                    finally:
                        results_queue.task_done()

            consumer_task = asyncio.create_task(_consume_results())

            if use_async:
                await connector.get_repos_with_stats_async(
                    org_name=org_name,
                    user_name=user_name,
                    pattern=pattern,
                    batch_size=batch_size,
                    max_concurrent=max_concurrent,
                    rate_limit_delay=rate_limit_delay,
                    max_commits_per_repo=max_commits_per_repo,
                    max_repos=max_repos,
                    on_repo_complete=on_repo_complete,
                )
            else:
                await loop.run_in_executor(
                    None,
                    lambda: connector.get_repos_with_stats(
                        org_name=org_name,
                        user_name=user_name,
                        pattern=pattern,
                        batch_size=batch_size,
                        max_concurrent=max_concurrent,
                        rate_limit_delay=rate_limit_delay,
                        max_commits_per_repo=max_commits_per_repo,
                        max_repos=max_repos,
                        on_repo_complete=on_repo_complete,
                    ),
                )

            await results_queue.join()
            await results_queue.put(_queue_sentinel)
            await consumer_task
        else:
            repos = await loop.run_in_executor(
                None,
                lambda: connector.list_repositories(
                    org_name=org_name,
                    user_name=user_name,
                    pattern=pattern,
                    max_repos=max_repos,
                ),
            )
            semaphore = asyncio.Semaphore(max(1, max_concurrent))

            async def _process_repo(repo_info) -> None:
                async with semaphore:
                    result = BatchResult(
                        repository=repo_info,
                        stats=None,
                        success=True,
                    )
                    try:
                        await store_result(result)
                    except Exception as e:
                        result = BatchResult(
                            repository=repo_info,
                            stats=None,
                            error=str(e),
                            success=False,
                        )
                    on_repo_complete(result)

            tasks = [asyncio.create_task(_process_repo(repo)) for repo in repos]
            if tasks:
                await asyncio.gather(*tasks)

        # Summary
        successful = sum(1 for r in all_results if r.success)
        failed = sum(1 for r in all_results if not r.success)
        logging.info("=== Batch Processing Complete ===")
        logging.info(f"  Successful: {successful}")
        logging.info(f"  Failed: {failed}")
        logging.info(f"  Total: {len(all_results)}")
        logging.info(f"  Stored: {stored_count}")

    except ConnectorException as e:
        logging.error(f"Connector error: {e}")
        raise
    except Exception as e:
        logging.error(f"Error in batch processing: {e}")
        raise
    finally:
        connector.close()



from datetime import date

import pytest
from fastapi.testclient import TestClient

from dev_health_ops.api.main import app
from dev_health_ops.api.models.schemas import (
    QuadrantAxes,
    QuadrantAxis,
    QuadrantPoint,
    QuadrantResponse,
)


def _validate(model, payload):
    if hasattr(model, "model_validate"):
        return model.model_validate(payload)
    return model.parse_obj(payload)


@pytest.fixture
def client():
    return TestClient(app)


def test_quadrant_endpoint_schema(client, monkeypatch):
    sample = QuadrantResponse(
        axes=QuadrantAxes(
            x=QuadrantAxis(metric="churn", label="Churn", unit="loc"),
            y=QuadrantAxis(metric="throughput", label="Throughput", unit="items"),
        ),
        points=[
            QuadrantPoint(
                entity_id="team-a",
                entity_label="Team A",
                x=42.0,
                y=12.0,
                window_start=date(2024, 1, 1),
                window_end=date(2024, 1, 8),
                evidence_link="/api/v1/explain?metric=throughput",
                trajectory=None,
            )
        ],
        annotations=[],
    )

    async def _fake_quadrant(**_):
        return sample

    monkeypatch.setattr("dev_health_ops.api.main.build_quadrant_response", _fake_quadrant)

    response = client.get("/api/v1/quadrant", params={"type": "churn_throughput"})
    assert response.status_code == 200
    _validate(QuadrantResponse, response.json())



import argparse
import asyncio
import logging
import os
import uuid
from datetime import date, datetime, time, timedelta, timezone

from dev_health_ops.work_graph.builder import BuildConfig, WorkGraphBuilder
from dev_health_ops.work_graph.investment.materialize import (
    MaterializeConfig,
    materialize_investments,
)


def _component_count(edges: list[tuple[str, str, str, str]]) -> int:
    """Count number of connected components in the graph."""
    adj: dict[tuple[str, str], list[tuple[str, str]]] = {}
    nodes: set[tuple[str, str]] = set()
    for source_type, source_id, target_type, target_id in edges:
        s = (source_type, source_id)
        t = (target_type, target_id)
        nodes.add(s)
        nodes.add(t)
        adj.setdefault(s, []).append(t)
        adj.setdefault(t, []).append(s)

    visited: set[tuple[str, str]] = set()
    count = 0
    for node in nodes:
        if node not in visited:
            count += 1
            stack = [node]
            visited.add(node)
            while stack:
                curr = stack.pop()
                for neighbor in adj.get(curr, []):
                    if neighbor not in visited:
                        visited.add(neighbor)
                        stack.append(neighbor)
    return count


def run_work_graph_build(ns: argparse.Namespace) -> int:
    """Build work graph edges from raw data."""

    # Parse dates
    from_date = None
    to_date = None

    if ns.from_date:
        from_date = datetime.fromisoformat(ns.from_date).replace(tzinfo=timezone.utc)
    else:
        from_date = datetime.now(timezone.utc) - timedelta(days=30)

    if ns.to_date:
        to_date = datetime.fromisoformat(ns.to_date).replace(tzinfo=timezone.utc)
    else:
        to_date = datetime.now(timezone.utc)

    # Parse repo_id if provided
    repo_id = None
    if ns.repo_id:
        repo_id = uuid.UUID(ns.repo_id)

    config = BuildConfig(
        dsn=ns.db,
        from_date=from_date,
        to_date=to_date,
        repo_id=repo_id,
        heuristic_days_window=ns.heuristic_window,
        heuristic_confidence=ns.heuristic_confidence,
    )

    logging.info(f"Building work graph from {config.from_date} to {config.to_date}")
    builder = WorkGraphBuilder(config)
    try:
        result = builder.build()

        total_edges = sum(result.values())
        logging.info("Work graph build complete. Total edges: %d", total_edges)
        logging.info("  issue_issue_edges: %d", result.get("issue_issue_edges", 0))
        logging.info("  issue_pr_edges: %d", result.get("issue_pr_edges", 0))
        logging.info("  pr_commit_edges: %d", result.get("pr_commit_edges", 0))
        logging.info("  commit_file_edges: %d", result.get("commit_file_edges", 0))
        logging.info("  heuristic_edges: %d", result.get("heuristic_edges", 0))

        client = getattr(builder, "client", None)
        if client is None:
            logging.error(
                "FAIL: Work graph builder did not expose a ClickHouse client."
            )
            return 1

        where_parts = [
            f"event_ts >= '{from_date.strftime('%Y-%m-%d %H:%M:%S')}'",
            f"event_ts <= '{to_date.strftime('%Y-%m-%d %H:%M:%S')}'",
        ]
        if repo_id:
            where_parts.append(f"repo_id = '{repo_id}'")
        where_sql = " AND ".join(where_parts)

        # Check edge count in DB for verification
        edge_count = client.query(
            f"SELECT count() FROM work_graph_edges WHERE {where_sql}"
        ).result_rows[0][0]
        if int(edge_count or 0) == 0:
            logging.error(
                "FAIL: work_graph_edges is empty for the selected window. "
                "Prerequisites missing or build produced no edges."
            )
            return 1

        if ns.check_components:
            edge_rows = (
                client.query(
                    f"""
                SELECT source_type, source_id, target_type, target_id
                FROM work_graph_edges
                WHERE {where_sql}
                """
                ).result_rows
                or []
            )
            edge_list = [
                (
                    str(e[0]),
                    str(e[1]),
                    str(e[2]),
                    str(e[3]),
                )
                for e in edge_rows
            ]
            comp_count = _component_count(edge_list)
            if comp_count == 1 and not getattr(ns, "allow_degenerate", False):
                logging.error(
                    "FAIL: Work graph is degenerate (connected_components=1). "
                    "Re-run with --allow-degenerate to override."
                )
                return 1
            logging.info("Connected components in window: %d", comp_count)

        return 0
    except Exception as e:
        logging.error(f"Work graph build failed: {e}")
        return 1
    finally:
        builder.close()


def run_investment_materialization(ns: argparse.Namespace) -> int:
    """Materialize investment metrics for a given window."""

    now = datetime.now(timezone.utc)
    if ns.to_date:
        to_day = date.fromisoformat(ns.to_date)
        to_ts = datetime.combine(
            to_day + timedelta(days=1), time.min, tzinfo=timezone.utc
        )
    else:
        to_ts = now

    if ns.from_date:
        from_day = date.fromisoformat(ns.from_date)
        from_ts = datetime.combine(from_day, time.min, tzinfo=timezone.utc)
    else:
        window_days = max(1, int(ns.window_days or 30))
        from_ts = to_ts - timedelta(days=window_days)

    if from_ts >= to_ts:
        logging.error("--from must be before --to")
        return 2

    repo_ids = [repo_id for repo_id in (ns.repo_id or []) if repo_id]
    team_ids = [team_id for team_id in (ns.team_id or []) if team_id]

    config = MaterializeConfig(
        dsn=ns.db,
        from_ts=from_ts,
        to_ts=to_ts,
        repo_ids=repo_ids or None,
        llm_provider=getattr(ns, "llm_provider", "auto") or "auto",
        persist_evidence_snippets=getattr(ns, "persist_evidence_snippets", False),
        llm_model=getattr(ns, "model", None),
        team_ids=team_ids or None,
        force=getattr(ns, "force", False),
    )

    logging.info(f"Materializing investments from {config.from_ts} to {config.to_ts}")
    try:
        stats = asyncio.run(materialize_investments(config))
        logging.info(
            "Investment materialization complete. Components=%d Records=%d Quotes=%d",
            stats.get("components", 0),
            stats.get("records", 0),
            stats.get("quotes", 0),
        )
        return 0
    except Exception as e:
        logging.error(f"Investment materialization failed: {e}")
        return 1


async def materialize_fixture_investments(
    *,
    db_url: str,
    from_ts: datetime,
    to_ts: datetime,
    repo_ids: list[str] | None = None,
    team_ids: list[str] | None = None,
) -> dict[str, int]:
    """Materialize fixture investments using the mock LLM provider."""
    config = MaterializeConfig(
        dsn=db_url,
        from_ts=from_ts,
        to_ts=to_ts,
        repo_ids=repo_ids,
        llm_provider="mock",
        persist_evidence_snippets=False,
        llm_model=None,
        team_ids=team_ids,
        force=True,
    )
    logging.info(
        "Materializing fixture investments from %s to %s",
        config.from_ts,
        config.to_ts,
    )
    return await materialize_investments(config)


def register_commands(subparsers: argparse._SubParsersAction) -> None:
    # ---- work-graph ----
    wg = subparsers.add_parser("work-graph", help="Work graph operations.")
    wg_sub = wg.add_subparsers(dest="work_graph_command", required=True)

    wg_build = wg_sub.add_parser("build", help="Build work graph edges from raw data.")
    wg_build.add_argument(
        "--db",
        required=True,
        help="ClickHouse connection string (clickhouse://user:pass@host:port/db).",
    )
    wg_build.add_argument(
        "--from",
        dest="from_date",
        type=str,
        help="Start date (YYYY-MM-DD). Defaults to 30 days ago.",
    )
    wg_build.add_argument(
        "--to",
        dest="to_date",
        type=str,
        help="End date (YYYY-MM-DD). Defaults to today.",
    )
    wg_build.add_argument(
        "--repo-id",
        type=str,
        help="Filter to specific repository UUID.",
    )
    wg_build.add_argument(
        "--heuristic-window",
        type=int,
        default=7,
        help="Days window for heuristic issue->PR matching (default: 7).",
    )
    wg_build.add_argument(
        "--heuristic-confidence",
        type=float,
        default=0.3,
        help="Confidence score for heuristic matches (default: 0.3).",
    )
    wg_build.add_argument(
        "--allow-degenerate",
        action="store_true",
        help="Allow single connected-component graphs (default: fail).",
    )
    wg_build.add_argument(
        "--check-components",
        action="store_true",
        default=True,
        help="Perform component analysis (enabled by default).",
    )
    wg_build.set_defaults(func=run_work_graph_build)

    # ---- investment ----
    investment = subparsers.add_parser(
        "investment", help="Investment materialization operations."
    )
    investment_sub = investment.add_subparsers(dest="investment_command", required=True)
    investment_materialize = investment_sub.add_parser(
        "materialize",
        help="Materialize work unit investment categorization into sinks.",
    )
    investment_materialize.add_argument(
        "--db",
        default=os.getenv("DATABASE_URI") or os.getenv("DATABASE_URL"),
        help="ClickHouse connection string (clickhouse://user:pass@host:port/db).",
    )
    investment_materialize.add_argument(
        "--from",
        dest="from_date",
        type=str,
        help="Start date (YYYY-MM-DD). Defaults to window-days before --to.",
    )
    investment_materialize.add_argument(
        "--to",
        dest="to_date",
        type=str,
        help="End date (YYYY-MM-DD). Defaults to now.",
    )
    investment_materialize.add_argument(
        "--window-days",
        type=int,
        default=30,
        help="Window size in days when --from is not set (default: 30).",
    )
    investment_materialize.add_argument(
        "--repo-id",
        action="append",
        default=[],
        help="Filter to specific repository UUID(s).",
    )
    investment_materialize.add_argument(
        "--team-id",
        action="append",
        default=[],
        help="Filter to specific team identifier(s).",
    )
    from dev_health_ops.llm.cli import add_llm_arguments

    add_llm_arguments(investment_materialize)
    investment_materialize.add_argument(
        "--persist-evidence-snippets",
        action="store_true",
        help="Persist extractive evidence quotes for work units.",
    )
    investment_materialize.add_argument(
        "--force", action="store_true", help="Force re-materialization."
    )
    investment_materialize.set_defaults(func=run_investment_materialization)



import asyncio
import logging
from datetime import datetime, timezone
from typing import Any, Iterable, List, Optional

from dev_health_ops.models.git import (
    GitCommit,
    GitCommitStat,
    Repo,
    GitPullRequest,
    GitBlame,
    GitFile,
    CiPipelineRun,
    Deployment,
    Incident,
)
from dev_health_ops.utils import (
    AGGREGATE_STATS_MARKER,
    is_skippable,
    CONNECTORS_AVAILABLE,
    BATCH_SIZE,
)

if CONNECTORS_AVAILABLE:
    from dev_health_ops.connectors import (
        BatchResult,
        GitLabConnector,
        ConnectorException,
    )
    from dev_health_ops.connectors.models import Repository
    from dev_health_ops.connectors.utils import RateLimitConfig, RateLimitGate
else:
    BatchResult = None  # type: ignore
    GitLabConnector = None  # type: ignore
    ConnectorException = Exception
    Repository = None  # type: ignore
    RateLimitConfig = None  # type: ignore
    RateLimitGate = None  # type: ignore


# --- GitLab Sync Helpers ---


def _fetch_gitlab_project_info_sync(connector, project_id):
    """Sync helper to fetch GitLab project info."""
    gl_project = connector.gitlab.projects.get(project_id)
    # Access properties to force load if lazy
    _ = gl_project.name
    return gl_project


def _fetch_gitlab_commits_sync(
    gl_project,
    max_commits: Optional[int],
    repo_id,
    since: Optional[datetime] = None,
):
    """Sync helper to fetch GitLab commits."""
    list_params = {"per_page": 100, "get_all": False}
    if since is not None:
        since_iso = since.astimezone(timezone.utc).isoformat().replace("+00:00", "Z")
        list_params["since"] = since_iso

    commit_objects = []
    count = 0
    commit_hashes = []
    page = 1
    per_page = min(max_commits, 100) if max_commits else 100
    stop_due_to_since = False

    while True:
        if max_commits is not None and count >= max_commits:
            break

        page_params = dict(list_params)
        page_params["page"] = page
        page_params["per_page"] = per_page
        commits_page = gl_project.commits.list(**page_params)
        if not commits_page:
            break
        logging.debug(
            "GitLab commits page %d returned %d items",
            page,
            len(commits_page),
        )

        for commit in commits_page:
            if max_commits is not None and count >= max_commits:
                break

            committed_when = None
            if hasattr(commit, "committed_date") and commit.committed_date:
                try:
                    committed_when = datetime.fromisoformat(
                        commit.committed_date.replace("Z", "+00:00")
                    )
                except Exception:
                    committed_when = None

            if since is not None and isinstance(committed_when, datetime):
                if committed_when.astimezone(timezone.utc) < since:
                    stop_due_to_since = True
                    break

            git_commit = GitCommit(
                repo_id=repo_id,
                hash=commit.id,
                message=commit.message,
                author_name=(
                    commit.author_name if hasattr(commit, "author_name") else "Unknown"
                ),
                author_email=None,
                author_when=(
                    datetime.fromisoformat(commit.authored_date.replace("Z", "+00:00"))
                    if hasattr(commit, "authored_date")
                    else datetime.now(timezone.utc)
                ),
                committer_name=(
                    commit.committer_name
                    if hasattr(commit, "committer_name")
                    else "Unknown"
                ),
                committer_email=None,
                committer_when=(
                    datetime.fromisoformat(commit.committed_date.replace("Z", "+00:00"))
                    if hasattr(commit, "committed_date")
                    else datetime.now(timezone.utc)
                ),
                parents=len(commit.parent_ids) if hasattr(commit, "parent_ids") else 0,
            )
            commit_objects.append(git_commit)
            commit_hashes.append(commit.id)
            count += 1

        if stop_due_to_since or len(commits_page) < per_page:
            break
        page += 1

    return commit_hashes, commit_objects


def _fetch_gitlab_commit_stats_sync(gl_project, commit_hashes, repo_id, max_stats):
    """Sync helper to fetch detailed commit stats from GitLab."""
    stats_objects = []

    for commit_hash in commit_hashes[:max_stats]:
        try:
            detailed_commit = gl_project.commits.get(commit_hash)
            if hasattr(detailed_commit, "stats"):
                stat = GitCommitStat(
                    repo_id=repo_id,
                    commit_hash=commit_hash,
                    file_path=AGGREGATE_STATS_MARKER,
                    additions=detailed_commit.stats.get("additions", 0),
                    deletions=detailed_commit.stats.get("deletions", 0),
                    old_file_mode="unknown",
                    new_file_mode="unknown",
                )
                stats_objects.append(stat)
        except Exception as e:
            logging.warning(
                "Failed to get stats for commit %s: %s",
                commit_hash,
                e,
            )
    return stats_objects


def _fetch_gitlab_mrs_sync(connector, project_id, repo_id, max_mrs):
    """Sync helper to fetch GitLab Merge Requests."""
    logging.info(
        "Fetching merge requests for project %d...",
        project_id,
    )
    mrs = connector.get_merge_requests(
        project_id=project_id, state="all", max_mrs=max_mrs
    )
    pr_objects = []
    for mr in mrs:
        created_at = (
            mr.created_at or mr.merged_at or mr.closed_at or datetime.now(timezone.utc)
        )
        git_pr = GitPullRequest(
            repo_id=repo_id,
            number=mr.number,
            title=mr.title,
            body=mr.body,
            state=mr.state,
            author_name=mr.author.username if mr.author else "Unknown",
            author_email=None,
            created_at=created_at,
            merged_at=mr.merged_at,
            closed_at=mr.closed_at,
            head_branch=mr.head_branch,
            base_branch=mr.base_branch,
        )
        pr_objects.append(git_pr)
    logging.info(
        "Fetched %d merge requests for project %d",
        len(pr_objects),
        project_id,
    )
    return pr_objects


def _sync_gitlab_mrs_to_store(
    connector,
    project_id: int,
    repo_id,
    store,
    loop: asyncio.AbstractEventLoop,
    batch_size: int,
    state: str = "all",
    gate: Optional[RateLimitGate] = None,
    since: Optional[datetime] = None,
) -> int:
    """Fetch all MRs for a project and insert them in batches.

    Runs in a worker thread; uses run_coroutine_threadsafe to write batches.
    """
    logging.info(
        "Fetching merge requests for project %d...",
        project_id,
    )
    batch: List[GitPullRequest] = []
    total = 0
    page = 1

    if gate is None:
        gate = RateLimitGate(RateLimitConfig(initial_backoff_seconds=1.0))

    while True:
        try:
            gate.wait_sync()
            logging.debug(
                "GitLab MRs page %d (per_page=%d) for project %d",
                page,
                connector.per_page,
                project_id,
            )
            mrs = connector.rest_client.get_merge_requests(
                project_id=project_id,
                state=state,
                page=page,
                per_page=connector.per_page,
                order_by="updated_at",
                sort="desc",
            )
            gate.reset()
        except Exception as e:
            retry_after = getattr(e, "retry_after_seconds", None)
            if retry_after is None:
                raise
            applied = gate.penalize(retry_after)
            logging.info(
                "GitLab rate limited while fetching MRs; backoff %.1fs (%s)",
                applied,
                e,
            )
            continue
        if not mrs:
            break
        logging.debug(
            "GitLab MRs page %d returned %d items (total: %d)",
            page,
            len(mrs),
            total,
        )

        for mr in mrs:
            author_name = "Unknown"
            author_email = None
            author_data = mr.get("author")
            if author_data:
                author_name = author_data.get("username") or author_name

            def _parse_dt(value):
                if not value:
                    return None
                try:
                    return datetime.fromisoformat(value.replace("Z", "+00:00"))
                except Exception:
                    return None

            created_at = _parse_dt(mr.get("created_at"))
            updated_at = _parse_dt(mr.get("updated_at"))
            merged_at = _parse_dt(mr.get("merged_at"))
            closed_at = _parse_dt(mr.get("closed_at"))
            created_at = (
                created_at or merged_at or closed_at or datetime.now(timezone.utc)
            )

            comments_count = int(mr.get("user_notes_count") or 0)

            if (
                since is not None
                and isinstance(updated_at, datetime)
                and updated_at.astimezone(timezone.utc) < since
            ):
                mrs = []
                break

            batch.append(
                GitPullRequest(
                    repo_id=repo_id,
                    number=int(mr.get("iid") or 0),
                    title=mr.get("title") or None,
                    body=mr.get("description"),
                    state=mr.get("state") or None,
                    author_name=author_name,
                    author_email=author_email,
                    created_at=created_at,
                    merged_at=merged_at,
                    closed_at=closed_at,
                    head_branch=mr.get("source_branch"),
                    base_branch=mr.get("target_branch"),
                    changes_requested_count=0,
                    reviews_count=0,
                    comments_count=comments_count,
                )
            )
            total += 1

            if len(batch) >= batch_size:
                asyncio.run_coroutine_threadsafe(
                    store.insert_git_pull_requests(batch),
                    loop,
                ).result()
                logging.debug(
                    "Stored batch of %d MRs for project %d (total: %d)",
                    len(batch),
                    project_id,
                    total,
                )
                batch.clear()

        page += 1
        if not mrs:
            break

    if batch:
        asyncio.run_coroutine_threadsafe(
            store.insert_git_pull_requests(batch),
            loop,
        ).result()

    logging.info(
        "Fetched %d merge requests for project %d",
        total,
        project_id,
    )
    return total


def _fetch_gitlab_pipelines_sync(gl_project, repo_id, max_pipelines, since):
    """Sync helper to fetch GitLab CI/CD pipelines."""
    pipelines = []

    def _coerce_datetime(value):
        if isinstance(value, datetime):
            return value
        if isinstance(value, str):
            try:
                return datetime.fromisoformat(value.replace("Z", "+00:00"))
            except Exception:
                return None
        return None

    try:
        list_params = {"per_page": 100, "order_by": "updated_at", "sort": "desc"}
        if max_pipelines > 100:
            raw_pipelines = gl_project.pipelines.list(**list_params, as_list=False)
        else:
            raw_pipelines = gl_project.pipelines.list(**list_params, get_all=False)
    except Exception as exc:
        logging.debug("Failed to fetch pipelines: %s", exc)
        return pipelines

    count = 0
    for pipeline in raw_pipelines:
        if count >= max_pipelines:
            break

        created_at = _coerce_datetime(getattr(pipeline, "created_at", None))

        if created_at is None:
            continue

        if since is not None and created_at.astimezone(timezone.utc) < since:
            break

        started_at = created_at
        started_at = (
            _coerce_datetime(getattr(pipeline, "started_at", None)) or created_at
        )

        finished_at = _coerce_datetime(getattr(pipeline, "finished_at", None))

        pipelines.append(
            CiPipelineRun(
                repo_id=repo_id,
                run_id=str(getattr(pipeline, "id", "")),
                status=getattr(pipeline, "status", None),
                queued_at=created_at,
                started_at=started_at,
                finished_at=finished_at,
            )
        )
        count += 1

    return pipelines


def _fetch_gitlab_deployments_sync(
    connector, project_id, repo_id, max_deployments, since
):
    """Sync helper to fetch GitLab deployments."""
    deployments = []
    try:
        # Use REST API to fetch deployments
        raw_deployments = connector.rest_client.get_deployments(
            project_id=project_id,
            per_page=min(max_deployments, 100),
            order_by="created_at",
            sort="desc",
        )
    except Exception as exc:
        logging.debug("Failed to fetch deployments: %s", exc)
        return deployments

    for dep in raw_deployments[:max_deployments]:
        created_at_str = dep.get("created_at")
        if not created_at_str:
            continue

        try:
            created_at = datetime.fromisoformat(created_at_str.replace("Z", "+00:00"))
        except Exception:
            continue

        if since is not None and created_at.astimezone(timezone.utc) < since:
            break

        # Parse other timestamps if available
        finished_at = None
        finished_at_str = dep.get("finished_at")
        if finished_at_str:
            try:
                finished_at = datetime.fromisoformat(
                    finished_at_str.replace("Z", "+00:00")
                )
            except Exception:
                pass

        deployments.append(
            Deployment(
                repo_id=repo_id,
                deployment_id=str(dep.get("id", "")),
                status=dep.get("status", None),
                environment=dep.get("environment", {}).get("name")
                if isinstance(dep.get("environment"), dict)
                else None,
                started_at=created_at,
                finished_at=finished_at,
                deployed_at=created_at,
                merged_at=None,
                pull_request_number=None,
            )
        )

    return deployments


def _fetch_gitlab_incidents_sync(connector, project_id, repo_id, max_issues, since):
    """Sync helper to fetch GitLab incidents (issues labeled 'incident')."""
    incidents = []
    try:
        raw_issues = connector.rest_client.get_issues(
            project_id=project_id,
            labels="incident",
            per_page=min(max_issues, 100),
            order_by="updated_at",
            sort="desc",
        )
    except Exception as exc:
        logging.debug("Failed to fetch incident issues: %s", exc)
        return incidents

    for issue in raw_issues[:max_issues]:
        created_at_str = issue.get("created_at")
        if not created_at_str:
            continue

        try:
            created_at = datetime.fromisoformat(created_at_str.replace("Z", "+00:00"))
        except Exception:
            continue

        if since is not None and created_at.astimezone(timezone.utc) < since:
            break

        resolved_at = None
        closed_at_str = issue.get("closed_at")
        if closed_at_str:
            try:
                resolved_at = datetime.fromisoformat(
                    closed_at_str.replace("Z", "+00:00")
                )
            except Exception:
                pass

        incidents.append(
            Incident(
                repo_id=repo_id,
                incident_id=str(issue.get("id", "")),
                status=issue.get("state", None),
                started_at=created_at,
                resolved_at=resolved_at,
            )
        )

    return incidents


def _iter_gitlab_repo_tree(
    gl_project,
    *,
    ref: str,
    per_page: int = 100,
    limit: Optional[int] = None,
) -> Iterable[Any]:
    page = 1
    seen = 0

    while True:
        try:
            page_items = gl_project.repository_tree(
                ref=ref,
                recursive=True,
                per_page=per_page,
                page=page,
                get_all=False,
            )
        except TypeError:
            page_items = gl_project.repository_tree(
                ref=ref,
                recursive=True,
                per_page=per_page,
                page=page,
                all=False,
            )
        if not page_items:
            break
        seen += len(page_items)
        logging.debug(
            "GitLab repo tree page %d returned %d items (total: %d)",
            page,
            len(page_items),
            seen,
        )
        for item in page_items:
            yield item
        if limit is not None and seen >= limit:
            return
        page += 1


def _fetch_gitlab_blame_sync(gl_project, connector, project_id, repo_id, limit=50):
    """Sync helper to fetch GitLab blame data."""
    blame_batch = []
    try:
        # Get files from repository tree (paged to avoid huge responses)
        files_to_process = []
        for item in _iter_gitlab_repo_tree(
            gl_project,
            ref=gl_project.default_branch,
            per_page=100,
            limit=None,
        ):
            if item["type"] == "blob" and not is_skippable(item["path"]):
                files_to_process.append(item["path"])
                if len(files_to_process) >= limit:
                    break

        # Limit files
        files_to_process = files_to_process[:limit]
        logging.debug(
            "GitLab blame: processing %d files (limit %d)",
            len(files_to_process),
            limit,
        )

        for idx, file_path in enumerate(files_to_process, start=1):
            try:
                logging.debug(
                    "GitLab blame fetch %d/%d: %s",
                    idx,
                    len(files_to_process),
                    file_path,
                )
                blame = connector.get_file_blame(
                    project_id=project_id,
                    file_path=file_path,
                    ref=gl_project.default_branch,
                )
                if blame and blame.ranges:
                    for r in blame.ranges:
                        blame_obj = GitBlame(
                            repo_id=repo_id,
                            path=file_path,
                            line_no=r.starting_line,
                            author_name=r.author,
                            author_email=r.author_email,
                            commit_hash=r.commit_sha,
                            line="<remote>",
                            author_when=datetime.now(timezone.utc),
                        )
                        blame_batch.append(blame_obj)
            except Exception as e:
                logging.warning(f"Failed to get blame for {file_path}: {e}")

    except Exception as e:
        logging.error(f"Error fetching files for blame: {e}")

    return blame_batch


async def _backfill_gitlab_missing_data(
    store: Any,
    connector: GitLabConnector,
    db_repo: Repo,
    project_full_name: str,
    default_branch: str,
    max_commits: Optional[int],
    blame_only: bool = False,
) -> None:
    if not (
        hasattr(store, "has_any_git_files")
        and hasattr(store, "has_any_git_blame")
        and hasattr(store, "has_any_git_commit_stats")
    ):
        return

    needs_files = not await store.has_any_git_files(db_repo.id)
    needs_commit_stats = (
        False if blame_only else not await store.has_any_git_commit_stats(db_repo.id)
    )
    needs_blame = not await store.has_any_git_blame(db_repo.id)

    if not (needs_files or needs_commit_stats or needs_blame):
        return

    try:
        project = connector.gitlab.projects.get(project_full_name)
    except Exception as e:
        logging.warning(f"Failed to load GitLab project {project_full_name}: {e}")
        return

    file_paths: List[str] = []
    if needs_files or needs_blame:
        try:
            items = _iter_gitlab_repo_tree(
                project,
                ref=default_branch,
                per_page=100,
                limit=None,
            )
        except Exception as e:
            logging.warning(f"Failed to list GitLab files for {project_full_name}: {e}")
            items = []

        for item in items or []:
            if item.get("type") != "blob":
                continue
            path = item.get("path")
            if not path:
                continue
            file_paths.append(path)

        if needs_files and file_paths:
            batch: List[GitFile] = []
            for path in file_paths:
                batch.append(
                    GitFile(
                        repo_id=db_repo.id,
                        path=path,
                        executable=False,
                        contents=None,
                    )
                )
                if len(batch) >= BATCH_SIZE:
                    await store.insert_git_file_data(batch)
                    batch.clear()
            if batch:
                await store.insert_git_file_data(batch)

    if needs_commit_stats:
        try:
            commits = project.commits.list(
                ref_name=default_branch, per_page=100, get_all=True
            )
        except TypeError:
            commits = project.commits.list(
                ref_name=default_branch, per_page=100, all=True
            )

        commit_stats_batch: List[GitCommitStat] = []
        commit_count = 0
        for commit in commits or []:
            if max_commits and commit_count >= max_commits:
                break
            commit_count += 1
            sha = getattr(commit, "id", None) or getattr(commit, "sha", None)
            if not sha:
                continue
            try:
                stats = connector.get_commit_stats_by_project(
                    sha=sha,
                    project_name=project_full_name,
                )
                commit_stats_batch.append(
                    GitCommitStat(
                        repo_id=db_repo.id,
                        commit_hash=sha,
                        file_path=AGGREGATE_STATS_MARKER,
                        additions=getattr(stats, "additions", 0),
                        deletions=getattr(stats, "deletions", 0),
                        old_file_mode="unknown",
                        new_file_mode="unknown",
                    )
                )
                if len(commit_stats_batch) >= BATCH_SIZE:
                    await store.insert_git_commit_stats(commit_stats_batch)
                    commit_stats_batch.clear()
            except Exception as e:
                logging.debug(
                    f"Failed commit stat fetch for {project_full_name}@{sha}: {e}"
                )

        if commit_stats_batch:
            await store.insert_git_commit_stats(commit_stats_batch)

    if needs_blame and file_paths:
        blame_batch: List[GitBlame] = []
        for path in file_paths:
            try:
                blame_items = connector.rest_client.get_file_blame(
                    project.id,
                    path,
                    default_branch,
                )
            except Exception as e:
                logging.debug(f"Failed blame fetch for {project_full_name}:{path}: {e}")
                continue

            line_no = 1
            for item in blame_items or []:
                commit = item.get("commit", {})
                for line in item.get("lines", []) or []:
                    blame_batch.append(
                        GitBlame(
                            repo_id=db_repo.id,
                            path=path,
                            line_no=line_no,
                            author_email=commit.get("author_email"),
                            author_name=commit.get("author_name"),
                            author_when=None,
                            commit_hash=commit.get("id"),
                            line=line.rstrip("\n") if isinstance(line, str) else None,
                        )
                    )
                    line_no += 1
                    if len(blame_batch) >= BATCH_SIZE:
                        await store.insert_blame_data(blame_batch)
                        blame_batch.clear()

        if blame_batch:
            await store.insert_blame_data(blame_batch)


async def process_gitlab_project(
    store: Any,
    project_id: int,
    token: str,
    gitlab_url: str,
    fetch_blame: bool = False,
    blame_only: bool = False,
    max_commits: Optional[int] = None,
    sync_git: bool = True,
    sync_prs: bool = True,
    sync_cicd: bool = True,
    sync_deployments: bool = True,
    sync_incidents: bool = True,
    since: Optional[datetime] = None,
) -> None:
    """
    Process a GitLab project using the GitLab connector.
    """
    if not CONNECTORS_AVAILABLE:
        raise RuntimeError(
            "Connectors are not available. Please install required dependencies."
        )

    logging.info(f"Processing GitLab project: {project_id}")
    loop = asyncio.get_running_loop()

    connector = GitLabConnector(url=gitlab_url, private_token=token)
    try:
        # 1. Fetch Project Info
        logging.info("Fetching project information...")
        gl_project = await loop.run_in_executor(
            None, _fetch_gitlab_project_info_sync, connector, project_id
        )

        logging.info(f"Found project: {gl_project.name}")

        # Create/Insert Repo
        full_name = (
            gl_project.path_with_namespace
            if hasattr(gl_project, "path_with_namespace")
            else gl_project.name
        )

        db_repo = Repo(
            repo_path=None,  # Not a local repo
            repo=full_name,
            settings={
                "source": "gitlab",
                "project_id": gl_project.id,
                "url": gl_project.web_url if hasattr(gl_project, "web_url") else None,
                "default_branch": (
                    gl_project.default_branch
                    if hasattr(gl_project, "default_branch")
                    else "main"
                ),
            },
            tags=["gitlab"],
        )

        await store.insert_repo(db_repo)
        logging.info(f"Project stored: {db_repo.repo} ({db_repo.id})")

        if blame_only:
            await _backfill_gitlab_missing_data(
                store=store,
                connector=connector,
                db_repo=db_repo,
                project_full_name=full_name,
                default_branch=db_repo.settings.get("default_branch", "main"),
                max_commits=max_commits,
                blame_only=True,
            )
            logging.info("Completed blame-only sync for GitLab project: %s", project_id)
            return

        if sync_git:
            # 2. Fetch Commits
            if max_commits is None:
                logging.info("Fetching all commits from GitLab...")
            else:
                logging.info(f"Fetching up to {max_commits} commits from GitLab...")
            commit_hashes, commit_objects = await loop.run_in_executor(
                None,
                _fetch_gitlab_commits_sync,
                gl_project,
                max_commits,
                db_repo.id,
                since,
            )

            if commit_objects:
                await store.insert_git_commit_data(commit_objects)
                logging.info(f"Stored {len(commit_objects)} commits from GitLab")

            # 3. Fetch Stats
            logging.info("Fetching commit stats from GitLab...")
            stats_limit = 50 if max_commits is None else min(max_commits, 50)
            stats_objects = await loop.run_in_executor(
                None,
                _fetch_gitlab_commit_stats_sync,
                gl_project,
                commit_hashes,
                db_repo.id,
                stats_limit,
            )

            if stats_objects:
                await store.insert_git_commit_stats(stats_objects)
                logging.info(f"Stored {len(stats_objects)} commit stats from GitLab")

        if sync_prs:
            # 4. Fetch Merge Requests
            logging.info("Fetching merge requests from GitLab...")
            mr_total = await loop.run_in_executor(
                None,
                _sync_gitlab_mrs_to_store,
                connector,
                project_id,
                db_repo.id,
                store,
                loop,
                BATCH_SIZE,
                "all",
                None,
                since,
            )
            logging.info(f"Stored {mr_total} merge requests from GitLab")

        if sync_cicd:
            logging.info("Fetching CI/CD pipelines from GitLab...")
            pipeline_runs = await loop.run_in_executor(
                None,
                _fetch_gitlab_pipelines_sync,
                gl_project,
                db_repo.id,
                BATCH_SIZE,
                since,
            )
            if pipeline_runs:
                await store.insert_ci_pipeline_runs(pipeline_runs)
                logging.info(f"Stored {len(pipeline_runs)} pipeline runs from GitLab")

        if sync_deployments:
            logging.info("Fetching deployments from GitLab...")
            deployments = await loop.run_in_executor(
                None,
                _fetch_gitlab_deployments_sync,
                connector,
                project_id,
                db_repo.id,
                BATCH_SIZE,
                since,
            )
            if deployments:
                await store.insert_deployments(deployments)
                logging.info(f"Stored {len(deployments)} deployments from GitLab")

        if sync_incidents:
            logging.info("Fetching incidents from GitLab...")
            incidents = await loop.run_in_executor(
                None,
                _fetch_gitlab_incidents_sync,
                connector,
                project_id,
                db_repo.id,
                BATCH_SIZE,
                since,
            )
            if incidents:
                await store.insert_incidents(incidents)
                logging.info(f"Stored {len(incidents)} incidents from GitLab")

        # 5. Fetch Blame (Optional)
        if fetch_blame:
            logging.info("Fetching blame data from GitLab (this may take a while)...")
            blame_batch = await loop.run_in_executor(
                None,
                _fetch_gitlab_blame_sync,
                gl_project,
                connector,
                project_id,
                db_repo.id,
                50,
            )

            if blame_batch:
                await store.insert_blame_data(blame_batch)
                logging.info(f"Stored {len(blame_batch)} blame records from GitLab")

        logging.info(f"Successfully processed GitLab project: {project_id}")

    except ConnectorException as e:
        logging.error(f"Connector error: {e}")
        raise
    except Exception as e:
        logging.error(f"Error processing GitLab project: {e}")
        raise
    finally:
        connector.close()


async def process_gitlab_projects_batch(
    store: Any,
    token: str,
    gitlab_url: str = "https://gitlab.com",
    group_name: str = None,
    pattern: str = None,
    batch_size: int = 10,
    max_concurrent: int = 4,
    rate_limit_delay: float = 1.0,
    max_commits_per_project: int = None,
    max_projects: int = None,
    use_async: bool = False,
    sync_git: bool = True,
    sync_prs: bool = True,
    sync_cicd: bool = True,
    sync_deployments: bool = True,
    sync_incidents: bool = True,
    blame_only: bool = False,
    backfill_missing: bool = True,
    since: Optional[datetime] = None,
) -> None:
    """
    Process multiple GitLab projects using batch processing with pattern matching.
    """
    if not CONNECTORS_AVAILABLE:
        raise RuntimeError(
            "Connectors are not available. Please install required dependencies."
        )

    logging.info("=== GitLab Batch Project Processing ===")
    connector = GitLabConnector(url=gitlab_url, private_token=token)
    loop = asyncio.get_running_loop()

    mr_gate = None
    mr_semaphore = None
    if sync_prs:
        mr_gate = RateLimitGate(
            RateLimitConfig(initial_backoff_seconds=max(1.0, rate_limit_delay))
        )
        mr_semaphore = asyncio.Semaphore(max(1, max_concurrent))

    all_results: List[BatchResult] = []
    stored_count = 0

    results_queue: Optional[asyncio.Queue] = None
    _queue_sentinel = object()

    async def store_result(result: BatchResult) -> None:
        """Store a single result in the database (upsert)."""
        nonlocal stored_count
        if not result.success:
            return

        project_info = result.repository
        db_repo = Repo(
            repo_path=None,  # Not a local repo
            repo=project_info.full_name,
            settings={
                "source": "gitlab",
                "project_id": project_info.id,
                "url": project_info.url,
                "default_branch": project_info.default_branch,
                "batch_processed": True,
            },
            tags=["gitlab"],
        )

        await store.insert_repo(db_repo)
        stored_count += 1
        logging.debug(f"Stored project ({stored_count}): {db_repo.repo}")

        if blame_only:
            try:
                await _backfill_gitlab_missing_data(
                    store=store,
                    connector=connector,
                    db_repo=db_repo,
                    project_full_name=project_info.full_name,
                    default_branch=project_info.default_branch,
                    max_commits=max_commits_per_project,
                    blame_only=True,
                )
            except Exception as e:
                logging.debug(
                    "Blame-only backfill failed for GitLab project %s: %s",
                    project_info.full_name,
                    e,
                )
            return

        gl_project = None
        if sync_git:
            # Fetch commits and stats to populate git_commits/git_commit_stats.
            if max_commits_per_project is None and since is None:
                commit_limit = 100
            else:
                commit_limit = max_commits_per_project
            try:
                if gl_project is None:
                    gl_project = await loop.run_in_executor(
                        None, connector.gitlab.projects.get, project_info.id
                    )
                commit_hashes, commit_objects = await loop.run_in_executor(
                    None,
                    _fetch_gitlab_commits_sync,
                    gl_project,
                    commit_limit,
                    db_repo.id,
                    since,
                )
                if commit_objects:
                    await store.insert_git_commit_data(commit_objects)

                stats_objects = await loop.run_in_executor(
                    None,
                    _fetch_gitlab_commit_stats_sync,
                    gl_project,
                    commit_hashes,
                    db_repo.id,
                    50 if commit_limit is None else min(commit_limit, 50),
                )
                if stats_objects:
                    await store.insert_git_commit_stats(stats_objects)
            except Exception as e:
                logging.warning(
                    "Failed to fetch commits for GitLab project %s: %s",
                    project_info.full_name,
                    e,
                )

        if sync_prs:
            # Fetch ALL merge requests for batch-processed projects, storing in batches.
            try:
                async with mr_semaphore:
                    await loop.run_in_executor(
                        None,
                        _sync_gitlab_mrs_to_store,
                        connector,
                        project_info.id,
                        db_repo.id,
                        store,
                        loop,
                        BATCH_SIZE,
                        "all",
                        mr_gate,
                        since,
                    )
            except Exception as e:
                logging.warning(
                    "Failed to fetch/store MRs for GitLab project %s: %s",
                    project_info.full_name,
                    e,
                )

        if sync_cicd:
            try:
                if gl_project is None:
                    gl_project = await loop.run_in_executor(
                        None, connector.gitlab.projects.get, project_info.id
                    )
                pipeline_runs = await loop.run_in_executor(
                    None,
                    _fetch_gitlab_pipelines_sync,
                    gl_project,
                    db_repo.id,
                    BATCH_SIZE,
                    since,
                )
                if pipeline_runs:
                    await store.insert_ci_pipeline_runs(pipeline_runs)
            except Exception as e:
                logging.warning(
                    "Failed to fetch CI/CD runs for GitLab project %s: %s",
                    project_info.full_name,
                    e,
                )

        if sync_deployments:
            try:
                deployments = await loop.run_in_executor(
                    None,
                    _fetch_gitlab_deployments_sync,
                    connector,
                    project_info.id,
                    db_repo.id,
                    BATCH_SIZE,
                    since,
                )
                if deployments:
                    await store.insert_deployments(deployments)
            except Exception as e:
                logging.warning(
                    "Failed to fetch deployments for GitLab project %s: %s",
                    project_info.full_name,
                    e,
                )

        if sync_incidents:
            try:
                incidents = await loop.run_in_executor(
                    None,
                    _fetch_gitlab_incidents_sync,
                    connector,
                    project_info.id,
                    db_repo.id,
                    BATCH_SIZE,
                    since,
                )
                if incidents:
                    await store.insert_incidents(incidents)
            except Exception as e:
                logging.warning(
                    "Failed to fetch incidents for GitLab project %s: %s",
                    project_info.full_name,
                    e,
                )

        if result.stats and sync_git:
            stat = GitCommitStat(
                repo_id=db_repo.id,
                commit_hash=AGGREGATE_STATS_MARKER,
                file_path=AGGREGATE_STATS_MARKER,
                additions=result.stats.additions,
                deletions=result.stats.deletions,
                old_file_mode="unknown",
                new_file_mode="unknown",
            )
            await store.insert_git_commit_stats([stat])

        if backfill_missing:
            try:
                await _backfill_gitlab_missing_data(
                    store=store,
                    connector=connector,
                    db_repo=db_repo,
                    project_full_name=project_info.full_name,
                    default_branch=project_info.default_branch,
                    max_commits=max_commits_per_project,
                )
            except Exception as e:
                logging.debug(
                    "Backfill failed for GitLab project %s: %s",
                    project_info.full_name,
                    e,
                )

    def on_project_complete(result: BatchResult) -> None:
        all_results.append(result)
        if result.success:
            stats_info = ""
            if result.stats:
                stats_info = f" ({result.stats.total_commits} commits)"
                logging.info(
                    f"  ✓ Processed: {result.repository.full_name}{stats_info}"
                )
        else:
            logging.warning(
                f"  ✗ Failed: {result.repository.full_name}: {result.error}"
            )

        if results_queue is not None:
            try:
                running_loop = asyncio.get_running_loop()
            except RuntimeError:
                running_loop = None

            def _enqueue() -> None:
                assert results_queue is not None
                try:
                    results_queue.put_nowait(result)
                except asyncio.QueueFull:
                    asyncio.create_task(results_queue.put(result))

            if running_loop is loop:
                _enqueue()
            else:
                loop.call_soon_threadsafe(_enqueue)

    try:
        if sync_git:
            results_queue = asyncio.Queue(maxsize=max(1, max_concurrent * 2))

            async def _consume_results() -> None:
                assert results_queue is not None
                while True:
                    item = await results_queue.get()
                    try:
                        if item is _queue_sentinel:
                            return
                        await store_result(item)
                    finally:
                        results_queue.task_done()

            consumer_task = asyncio.create_task(_consume_results())

            if use_async:
                await connector.get_projects_with_stats_async(
                    group_name=group_name,
                    pattern=pattern,
                    batch_size=batch_size,
                    max_concurrent=max_concurrent,
                    rate_limit_delay=rate_limit_delay,
                    max_commits_per_repo=max_commits_per_project,
                    max_repos=max_projects,
                    on_project_complete=on_project_complete,
                )
            else:
                await loop.run_in_executor(
                    None,
                    lambda: connector.get_projects_with_stats(
                        group_name=group_name,
                        pattern=pattern,
                        batch_size=batch_size,
                        max_concurrent=max_concurrent,
                        rate_limit_delay=rate_limit_delay,
                        max_commits_per_repo=max_commits_per_project,
                        max_repos=max_projects,
                        on_project_complete=on_project_complete,
                    ),
                )

            await results_queue.join()
            await results_queue.put(_queue_sentinel)
            await consumer_task
        else:
            logging.info(
                "Listing GitLab projects for PR sync (group=%s, pattern=%s, max=%s)",
                group_name,
                pattern,
                max_projects,
            )
            projects = await loop.run_in_executor(
                None,
                lambda: connector._get_projects_for_processing(
                    group_name=group_name,
                    pattern=pattern,
                    max_repos=max_projects,
                ),
            )
            logging.info("Discovered %d GitLab projects for PR sync", len(projects))
            semaphore = asyncio.Semaphore(max(1, max_concurrent))

            async def _process_project(project_info) -> None:
                async with semaphore:
                    result = BatchResult(
                        repository=project_info,
                        stats=None,
                        success=True,
                    )
                    try:
                        await store_result(result)
                    except Exception as e:
                        result = BatchResult(
                            repository=project_info,
                            stats=None,
                            error=str(e),
                            success=False,
                        )
                    on_project_complete(result)

            tasks = [asyncio.create_task(_process_project(p)) for p in projects]
            if tasks:
                await asyncio.gather(*tasks)

        # Summary
        successful = sum(1 for r in all_results if r.success)
        failed = sum(1 for r in all_results if not r.success)
        logging.info("=== Batch Processing Complete ===")
        logging.info(f"  Successful: {successful}")
        logging.info(f"  Failed: {failed}")
        logging.info(f"  Total: {len(all_results)}")
        logging.info(f"  Stored: {stored_count}")

    except ConnectorException as e:
        logging.error(f"Connector error: {e}")
        raise
    except Exception as e:
        logging.error(f"Error in batch processing: {e}")
        raise
    finally:
        connector.close()



from __future__ import annotations

from datetime import date, datetime
from typing import Any, Dict, List

from .client import query_dicts


async def fetch_investment_flow_items(
    client: Any,
    *,
    start_ts: datetime,
    end_ts: datetime,
    scope_filter: str,
    scope_params: Dict[str, Any],
    limit: int,
) -> List[Dict[str, Any]]:
    query = f"""
        SELECT
            theme_kv.1 AS source,
            ifNull(r.repo, toString(repo_id)) AS target,
            sum(theme_kv.2 * effort_value) AS value
        FROM work_unit_investments
        LEFT JOIN repos AS r ON r.id = repo_id
        ARRAY JOIN CAST(theme_distribution_json AS Array(Tuple(String, Float32))) AS theme_kv
        WHERE from_ts < %(end_ts)s AND to_ts >= %(start_ts)s
            {scope_filter}
        GROUP BY source, target
        ORDER BY value DESC
        LIMIT %(limit)s
    """
    params = {"start_ts": start_ts, "end_ts": end_ts, "limit": limit}
    params.update(scope_params)
    return await query_dicts(client, query, params)


async def fetch_expense_counts(
    client: Any,
    *,
    start_day: date,
    end_day: date,
    scope_filter: str,
    scope_params: Dict[str, Any],
) -> List[Dict[str, Any]]:
    query = f"""
        SELECT
            sum(new_items_count) AS new_items,
            sum(new_bugs_count) AS new_bugs,
            sum(items_completed * bug_completed_ratio) AS bug_completed_estimate
        FROM work_item_metrics_daily
        WHERE day >= %(start_day)s AND day < %(end_day)s
            {scope_filter}
    """
    params = {"start_day": start_day, "end_day": end_day}
    params.update(scope_params)
    return await query_dicts(client, query, params)


async def fetch_expense_abandoned(
    client: Any,
    *,
    start_day: date,
    end_day: date,
    scope_filter: str,
    scope_params: Dict[str, Any],
) -> List[Dict[str, Any]]:
    query = f"""
        SELECT
            countIf(status = 'canceled') AS canceled_items
        FROM work_item_cycle_times
        WHERE day >= %(start_day)s AND day < %(end_day)s
            {scope_filter}
    """
    params = {"start_day": start_day, "end_day": end_day}
    params.update(scope_params)
    return await query_dicts(client, query, params)


async def fetch_state_status_counts(
    client: Any,
    *,
    start_day: date,
    end_day: date,
    scope_filter: str,
    scope_params: Dict[str, Any],
) -> List[Dict[str, Any]]:
    query = f"""
        SELECT
            status,
            sum(items_touched) AS items_touched
        FROM work_item_state_durations_daily
        WHERE day >= %(start_day)s AND day < %(end_day)s
            {scope_filter}
        GROUP BY status
        ORDER BY items_touched DESC
    """
    params = {"start_day": start_day, "end_day": end_day}
    params.update(scope_params)
    return await query_dicts(client, query, params)


async def fetch_hotspot_rows(
    client: Any,
    *,
    start_day: date,
    end_day: date,
    scope_filter: str,
    scope_params: Dict[str, Any],
    limit: int,
) -> List[Dict[str, Any]]:
    query = f"""
        WITH
            (
                SELECT quantileExact(0.7)(churn) FROM (
                    SELECT
                        sum(metrics.churn) AS churn
                    FROM file_metrics_daily AS metrics
                    INNER JOIN repos AS r ON r.id = metrics.repo_id
                    WHERE metrics.day >= %(start_day)s AND metrics.day < %(end_day)s
                        {scope_filter}
                    GROUP BY r.repo, metrics.path
                )
            ) AS churn_hi,
            (
                SELECT quantileExact(0.3)(churn) FROM (
                    SELECT
                        sum(metrics.churn) AS churn
                    FROM file_metrics_daily AS metrics
                    INNER JOIN repos AS r ON r.id = metrics.repo_id
                    WHERE metrics.day >= %(start_day)s AND metrics.day < %(end_day)s
                        {scope_filter}
                    GROUP BY r.repo, metrics.path
                )
            ) AS churn_mid
        SELECT
            r.repo AS repo,
            if(
                position(metrics.path, '/') > 0,
                arrayElement(splitByChar('/', metrics.path), 1),
                '(root)'
            ) AS directory,
            metrics.path AS file_path,
            multiIf(
                sum(metrics.churn) >= churn_hi,
                'refactor',
                sum(metrics.churn) >= churn_mid,
                'fix',
                'feature'
            ) AS change_type,
            sum(metrics.churn) AS churn
        FROM file_metrics_daily AS metrics
        INNER JOIN repos AS r
            ON r.id = metrics.repo_id
        WHERE metrics.day >= %(start_day)s AND metrics.day < %(end_day)s
            AND metrics.path != ''
            {scope_filter}
        GROUP BY repo, directory, file_path
        ORDER BY churn DESC
        LIMIT %(limit)s
    """
    params = {"start_day": start_day, "end_day": end_day, "limit": limit}
    params.update(scope_params)
    return await query_dicts(client, query, params)



import logging
import uuid
from datetime import date, datetime, timedelta, timezone
from pathlib import Path
from typing import Optional, Any

import git
from dev_health_ops.analytics.complexity import ComplexityScanner
from dev_health_ops.metrics.schemas import FileComplexitySnapshot, RepoComplexityDaily
from dev_health_ops.metrics.sinks.clickhouse import ClickHouseMetricsSink
from dev_health_ops.metrics.sinks.sqlite import SQLiteMetricsSink
from dev_health_ops.metrics.sinks.mongo import MongoMetricsSink
from dev_health_ops.storage import detect_db_type

logger = logging.getLogger(__name__)

DEFAULT_COMPLEXITY_CONFIG_PATH = Path(__file__).resolve().parents[1] / "config" / "complexity.yaml"


def _normalize_sqlite_url(db_url: str) -> str:
    if "sqlite+aiosqlite://" in db_url:
        return db_url.replace("sqlite+aiosqlite://", "sqlite://", 1)
    return db_url


def _normalize_postgres_url(db_url: str) -> str:
    if "postgresql+asyncpg://" in db_url:
        return db_url.replace("postgresql+asyncpg://", "postgresql://", 1)
    return db_url


def run_complexity_scan_job(
    *,
    repo_path: Path,
    repo_id: uuid.UUID,
    db_url: str,
    date: date,
    backfill_days: int = 1,
    ref: str = "HEAD",
    sink: Optional[Any] = None,
) -> None:
    if not db_url and not sink:
        raise ValueError("DB connection string or sink is required.")

    if not repo_path.exists():
        raise FileNotFoundError(f"Repo path {repo_path} does not exist.")

    repo = git.Repo(repo_path)
    scanner = ComplexityScanner(config_path=DEFAULT_COMPLEXITY_CONFIG_PATH)

    # Calculate date range
    start_date = date - timedelta(days=max(1, backfill_days) - 1)
    dates_to_process = [
        start_date + timedelta(days=i) for i in range(max(1, backfill_days))
    ]

    own_sink = False
    if sink is None:
        backend = detect_db_type(db_url)
        own_sink = True
        if backend == "clickhouse":
            sink = ClickHouseMetricsSink(db_url)
        elif backend == "sqlite":
            sink = SQLiteMetricsSink(_normalize_sqlite_url(db_url))
        elif backend == "mongo":
            sink = MongoMetricsSink(db_url)
        elif backend == "postgres":
            sink = SQLiteMetricsSink(_normalize_postgres_url(db_url))
        else:
            raise ValueError(f"Unsupported backend for complexity job: {backend}")

    try:
        if own_sink:
            sink.ensure_tables()

        for d in dates_to_process:
            logger.info(f"Processing date: {d}")

            # Find commit for this day
            # "last commit before end of day"
            # git rev-list -1 --before="YYYY-MM-DD 23:59:59" ref
            end_of_day_ts = (
                datetime.combine(d, datetime.max.time())
                .replace(tzinfo=timezone.utc)
                .isoformat()
            )

            try:
                commit_sha = repo.git.rev_list("-1", f"--before={end_of_day_ts}", ref)
            except git.Exc as e:
                logger.warning(f"Could not find commit for {d}: {e}")
                continue

            if not commit_sha:
                logger.warning(f"No commit found before {d} (ref={ref}). Skipping.")
                continue

            commit_sha = commit_sha.strip()
            logger.info(f"Scanning {repo_path} at commit {commit_sha} (as of {d})...")

            file_results = scanner.scan_git_ref(repo_path, commit_sha)
            logger.info(f"Scanned {len(file_results)} files.")

            if not file_results:
                logger.warning(f"No complexity data found for {d}.")
                continue

            # Aggregations
            computed_at = datetime.now(timezone.utc)

            snapshots = []
            total_loc = 0
            total_cc = 0
            total_high = 0
            total_very_high = 0

            for f in file_results:
                snapshots.append(
                    FileComplexitySnapshot(
                        repo_id=repo_id,
                        as_of_day=d,
                        ref=commit_sha,
                        file_path=f.file_path,
                        language=f.language,
                        loc=f.loc,
                        functions_count=f.functions_count,
                        cyclomatic_total=f.cyclomatic_total,
                        cyclomatic_avg=f.cyclomatic_avg,
                        high_complexity_functions=f.high_complexity_functions,
                        very_high_complexity_functions=f.very_high_complexity_functions,
                        computed_at=computed_at,
                    )
                )

                total_loc += f.loc
                total_cc += f.cyclomatic_total
                total_high += f.high_complexity_functions
                total_very_high += f.very_high_complexity_functions

            cc_per_kloc = (total_cc / (total_loc / 1000.0)) if total_loc > 0 else 0.0

            repo_daily = RepoComplexityDaily(
                repo_id=repo_id,
                day=d,
                loc_total=total_loc,
                cyclomatic_total=total_cc,
                cyclomatic_per_kloc=cc_per_kloc,
                high_complexity_functions=total_high,
                very_high_complexity_functions=total_very_high,
                computed_at=computed_at,
            )

            logger.info(f"Writing {len(snapshots)} file snapshots for {d}...")
            sink.write_file_complexity_snapshots(snapshots)
            logger.info(f"Writing repo daily summary for {d}...")
            sink.write_repo_complexity_daily([repo_daily])

    finally:
        if own_sink:
            sink.close()

    logger.info("Complexity job done.")



from __future__ import annotations

import logging
import re
from datetime import datetime, timezone
from typing import Any, List, Optional, Sequence, Tuple

from dev_health_ops.models.work_items import (
    Sprint,
    WorkItem,
    WorkItemDependency,
    WorkItemInteractionEvent,
    WorkItemReopenEvent,
    WorkItemStatusTransition,
)
from dev_health_ops.providers.identity import IdentityResolver
from dev_health_ops.providers.status_mapping import StatusMapping

logger = logging.getLogger(__name__)

# Priority labels -> (priority_raw, service_class)
_PRIORITY_LABEL_MAP = {
    "priority::critical": ("critical", "expedite"),
    "priority::high": ("high", "fixed_date"),
    "priority::medium": ("medium", "standard"),
    "priority::low": ("low", "intangible"),
    "critical": ("critical", "expedite"),
    "blocker": ("critical", "expedite"),
    "high": ("high", "fixed_date"),
    "medium": ("medium", "standard"),
    "low": ("low", "intangible"),
    "p0": ("critical", "expedite"),
    "p1": ("high", "fixed_date"),
    "p2": ("medium", "standard"),
    "p3": ("low", "intangible"),
    "p4": ("low", "intangible"),
}


def _parse_iso(value: Optional[str]) -> Optional[datetime]:
    if not value:
        return None
    try:
        return datetime.fromisoformat(str(value).replace("Z", "+00:00"))
    except Exception:
        return None


def _to_utc(dt: Optional[datetime]) -> Optional[datetime]:
    if dt is None:
        return None
    if dt.tzinfo is None:
        return dt.replace(tzinfo=timezone.utc)
    return dt.astimezone(timezone.utc)


def _priority_from_labels(labels: Sequence[str]) -> Tuple[Optional[str], Optional[str]]:
    """
    Extract priority_raw and service_class from GitLab labels.

    Returns (priority_raw, service_class) or (None, None) if no match.
    """
    for label in labels:
        key = label.lower().strip()
        if key in _PRIORITY_LABEL_MAP:
            return _PRIORITY_LABEL_MAP[key]
    return (None, None)


def _get(obj: Any, key: str) -> Any:
    if isinstance(obj, dict):
        return obj.get(key)
    return getattr(obj, key, None)


def gitlab_issue_to_work_item(
    *,
    issue: Any,
    project_full_path: str,
    repo_id: Optional[Any],
    status_mapping: StatusMapping,
    identity: IdentityResolver,
    label_events: Optional[Sequence[Any]] = None,
) -> Tuple[WorkItem, List[WorkItemStatusTransition]]:
    iid = int(_get(issue, "iid") or 0)
    work_item_id = f"gitlab:{project_full_path}#{iid}"

    title = _get(issue, "title") or ""
    description = _get(issue, "description")
    state = _get(issue, "state") or None  # opened/closed
    created_at = _to_utc(_parse_iso(_get(issue, "created_at"))) or datetime.now(
        timezone.utc
    )
    updated_at = _to_utc(_parse_iso(_get(issue, "updated_at"))) or created_at
    closed_at = _to_utc(_parse_iso(_get(issue, "closed_at")))

    labels = list(_get(issue, "labels") or [])
    labels = [str(lbl) for lbl in labels if lbl]

    normalized_status = status_mapping.normalize_status(
        provider="gitlab",
        status_raw=None,
        labels=labels,
        state=str(state) if state else None,
    )
    normalized_type = status_mapping.normalize_type(
        provider="gitlab",
        type_raw=None,
        labels=labels,
    )

    assignees: List[str] = []
    for a in _get(issue, "assignees") or []:
        assignees.append(
            identity.resolve(
                provider="gitlab",
                email=_get(a, "email"),
                username=_get(a, "username"),
                display_name=_get(a, "name"),
            )
        )

    author_obj = _get(issue, "author")
    reporter = None
    if author_obj is not None:
        reporter = identity.resolve(
            provider="gitlab",
            email=_get(author_obj, "email"),
            username=_get(author_obj, "username"),
            display_name=_get(author_obj, "name"),
        )

    url = _get(issue, "web_url") or _get(issue, "url")

    # Best-effort transitions from label events + state.
    transitions: List[WorkItemStatusTransition] = []
    started_at = None
    completed_at = None

    if label_events:

        def _ev_dt(ev: Any) -> datetime:
            return _to_utc(_parse_iso(_get(ev, "created_at"))) or datetime.min.replace(
                tzinfo=timezone.utc
            )

        prev_status = "unknown"
        for ev in sorted(list(label_events), key=_ev_dt):
            action = str(_get(ev, "action") or "").lower()
            label = _get(ev, "label") or {}
            label_name = _get(label, "name") or _get(ev, "label_name")
            if not label_name:
                continue
            label_name = str(label_name)
            occurred_at = _to_utc(_parse_iso(_get(ev, "created_at"))) or created_at

            if action not in {"add", "remove"}:
                continue
            mapped = status_mapping.normalize_status(
                provider="gitlab",
                status_raw=None,
                labels=[label_name] if action == "add" else (),
                state=None,
            )
            if mapped == "unknown":
                continue
            transitions.append(
                WorkItemStatusTransition(
                    work_item_id=work_item_id,
                    provider="gitlab",
                    occurred_at=occurred_at,
                    from_status_raw=None,
                    to_status_raw=label_name,
                    from_status=prev_status,  # type: ignore[arg-type]
                    to_status=mapped,
                    actor=None,
                )
            )
            prev_status = mapped

        for t in transitions:
            if started_at is None and t.to_status == "in_progress":
                started_at = t.occurred_at
            if completed_at is None and t.to_status in {"done", "canceled"}:
                completed_at = t.occurred_at
                break

    if completed_at is None and closed_at is not None:
        completed_at = closed_at

    weight = _get(issue, "weight")
    story_points = float(weight) if weight is not None else None

    work_item = WorkItem(
        work_item_id=work_item_id,
        provider="gitlab",
        repo_id=repo_id,
        project_key=None,
        # For work tracking metrics, treat the GitLab project path as the "project" scope.
        project_id=str(project_full_path)
        if project_full_path
        else (str(_get(issue, "project_id")) if _get(issue, "project_id") else None),
        title=str(title),
        description=str(description) if description else None,
        type=normalized_type,
        status=normalized_status,
        status_raw=str(state) if state else None,
        assignees=[a for a in assignees if a and a != "unknown"],
        reporter=reporter if reporter and reporter != "unknown" else None,
        created_at=created_at,
        updated_at=updated_at,
        started_at=started_at,
        completed_at=completed_at,
        closed_at=closed_at,
        labels=labels,
        story_points=story_points,
        sprint_id=str(_get(_get(issue, "milestone"), "id"))
        if _get(issue, "milestone")
        else None,
        sprint_name=_get(_get(issue, "milestone"), "title")
        if _get(issue, "milestone")
        else None,
        url=url,
    )
    return work_item, transitions


def gitlab_mr_to_work_item(
    *,
    mr: Any,
    project_full_path: str,
    repo_id: Optional[Any],
    status_mapping: StatusMapping,
    identity: IdentityResolver,
    state_events: Optional[Sequence[Any]] = None,
) -> Tuple[WorkItem, List[WorkItemStatusTransition]]:
    """
    Convert a GitLab merge request to a normalized WorkItem.

    MRs are treated as work items with type "merge_request".
    State events are used for transitions (opened->merged/closed).
    """
    iid = int(_get(mr, "iid") or 0)
    work_item_id = f"gitlab:{project_full_path}!{iid}"  # ! for MRs

    title = _get(mr, "title") or ""
    description = _get(mr, "description")
    state = _get(mr, "state") or None  # opened/merged/closed
    created_at = _to_utc(_parse_iso(_get(mr, "created_at"))) or datetime.now(
        timezone.utc
    )
    updated_at = _to_utc(_parse_iso(_get(mr, "updated_at"))) or created_at
    merged_at = _to_utc(_parse_iso(_get(mr, "merged_at")))
    closed_at = _to_utc(_parse_iso(_get(mr, "closed_at")))

    labels = list(_get(mr, "labels") or [])
    labels = [str(lb) for lb in labels if lb]

    # MRs use state-based status
    status_raw = str(state) if state else "unknown"
    if status_raw == "merged":
        normalized_status = "done"
    elif status_raw == "closed":
        normalized_status = "canceled"
    elif status_raw == "opened":
        normalized_status = "in_progress"
    else:
        normalized_status = "unknown"

    # Priority from labels
    priority_raw, service_class = _priority_from_labels(labels)

    assignees: List[str] = []
    for a in _get(mr, "assignees") or []:
        assignees.append(
            identity.resolve(
                provider="gitlab",
                email=_get(a, "email"),
                username=_get(a, "username"),
                display_name=_get(a, "name"),
            )
        )

    author_obj = _get(mr, "author")
    reporter = None
    if author_obj is not None:
        reporter = identity.resolve(
            provider="gitlab",
            email=_get(author_obj, "email"),
            username=_get(author_obj, "username"),
            display_name=_get(author_obj, "name"),
        )

    url = _get(mr, "web_url") or _get(mr, "url")

    # Transitions from state events
    transitions: List[WorkItemStatusTransition] = []
    started_at = created_at  # MRs start when opened
    completed_at = merged_at or closed_at

    if state_events:
        prev_status = "unknown"
        for ev in sorted(
            state_events,
            key=lambda e: (
                _to_utc(_parse_iso(_get(e, "created_at")))
                or datetime.min.replace(tzinfo=timezone.utc)
            ),
        ):
            ev_state = str(_get(ev, "state") or "").lower()
            occurred_at = _to_utc(_parse_iso(_get(ev, "created_at"))) or created_at

            if ev_state == "merged":
                to_status = "done"
            elif ev_state == "closed":
                to_status = "canceled"
            elif ev_state == "opened":
                to_status = "in_progress"
            elif ev_state == "reopened":
                to_status = "in_progress"
            else:
                continue

            user_obj = _get(ev, "user")
            actor = None
            if user_obj:
                actor = identity.resolve(
                    provider="gitlab",
                    email=_get(user_obj, "email"),
                    username=_get(user_obj, "username"),
                    display_name=_get(user_obj, "name"),
                )

            transitions.append(
                WorkItemStatusTransition(
                    work_item_id=work_item_id,
                    provider="gitlab",
                    occurred_at=occurred_at,
                    from_status_raw=None,
                    to_status_raw=ev_state,
                    from_status=prev_status,
                    to_status=to_status,
                    actor=actor,
                )
            )
            prev_status = to_status

    weight = _get(mr, "weight")
    story_points = float(weight) if weight is not None else None

    work_item = WorkItem(
        work_item_id=work_item_id,
        provider="gitlab",
        repo_id=repo_id,
        project_key=None,
        project_id=str(project_full_path) if project_full_path else None,
        title=str(title),
        description=str(description) if description else None,
        type="merge_request",
        status=normalized_status,
        status_raw=status_raw,
        assignees=[a for a in assignees if a and a != "unknown"],
        reporter=reporter if reporter and reporter != "unknown" else None,
        created_at=created_at,
        updated_at=updated_at,
        started_at=started_at,
        completed_at=completed_at,
        closed_at=closed_at or merged_at,
        labels=labels,
        story_points=story_points,
        sprint_id=str(_get(_get(mr, "milestone"), "id"))
        if _get(mr, "milestone")
        else None,
        sprint_name=_get(_get(mr, "milestone"), "title")
        if _get(mr, "milestone")
        else None,
        url=url,
        priority_raw=priority_raw,
        service_class=service_class,
    )
    return work_item, transitions


def detect_gitlab_reopen_events(
    *,
    work_item_id: str,
    state_events: Sequence[Any],
    identity: IdentityResolver,
) -> List[WorkItemReopenEvent]:
    """
    Detect reopen events from GitLab resource_state_events.

    A reopen occurs when state changes to "reopened".
    """
    events: List[WorkItemReopenEvent] = []
    for ev in state_events:
        ev_state = str(_get(ev, "state") or "").lower()
        if ev_state != "reopened":
            continue

        occurred_at = _to_utc(_parse_iso(_get(ev, "created_at")))
        if not occurred_at:
            continue

        user_obj = _get(ev, "user")
        actor = None
        if user_obj:
            actor = identity.resolve(
                provider="gitlab",
                email=_get(user_obj, "email"),
                username=_get(user_obj, "username"),
                display_name=_get(user_obj, "name"),
            )

        events.append(
            WorkItemReopenEvent(
                work_item_id=work_item_id,
                occurred_at=occurred_at,
                from_status="done",  # Reopen implies was closed/done
                to_status="in_progress",
                from_status_raw="closed",
                to_status_raw="reopened",
                actor=actor,
            )
        )
    return events


def gitlab_note_to_interaction_event(
    *,
    note: Any,
    work_item_id: str,
    identity: IdentityResolver,
) -> Optional[WorkItemInteractionEvent]:
    """
    Convert a GitLab note (comment/discussion) to an interaction event.

    System notes are excluded unless they indicate meaningful work.
    """
    if _get(note, "system"):
        # Skip system-generated notes (label changes, assignments, etc.)
        return None

    body = _get(note, "body") or ""
    occurred_at = _to_utc(_parse_iso(_get(note, "created_at")))
    if not occurred_at:
        return None

    author_obj = _get(note, "author")
    actor = None
    if author_obj:
        actor = identity.resolve(
            provider="gitlab",
            email=_get(author_obj, "email"),
            username=_get(author_obj, "username"),
            display_name=_get(author_obj, "name"),
        )

    return WorkItemInteractionEvent(
        work_item_id=work_item_id,
        provider="gitlab",
        interaction_type="comment",
        occurred_at=occurred_at,
        actor=actor,
        body_length=len(body),
    )


# Regex patterns for GitLab issue references
_GITLAB_ISSUE_REF_PATTERN = re.compile(
    r"(?:^|[^\w])(?:(?P<project>[\w/-]+)?#(?P<iid>\d+))",
    re.MULTILINE,
)
_BLOCKING_KEYWORDS = {"blocks", "blocked by", "is blocked by", "blocking"}


def extract_gitlab_dependencies(
    *,
    work_item_id: str,
    issue: Any,
    project_full_path: str,
    linked_issues: Optional[Sequence[Any]] = None,
) -> List[WorkItemDependency]:
    """
    Extract dependency edges from GitLab issue links and description.

    GitLab has explicit issue links (via API) and implicit references in description.
    """
    dependencies: List[WorkItemDependency] = []
    seen_targets: set[str] = set()

    # Process explicit links from API
    if linked_issues:
        for link in linked_issues:
            link_type = str(_get(link, "link_type") or "relates_to").lower()
            target_iid = _get(link, "iid")

            if not target_iid:
                continue

            # Build target work_item_id
            # Linked issues might be from same or different project
            refs = _get(link, "references")
            target_path = (
                _get(refs, "full") if isinstance(refs, dict) else None
            ) or project_full_path
            if target_path and "#" in str(target_path):
                # Extract project from "group/project#123" format
                target_path = str(target_path).split("#")[0]

            target_id = f"gitlab:{target_path}#{target_iid}"
            if target_id in seen_targets:
                continue
            seen_targets.add(target_id)

            # Map GitLab link types
            if link_type in {"blocks", "is_blocked_by"}:
                relationship = "blocks" if link_type == "blocks" else "blocked_by"
            else:
                relationship = "relates_to"

            dependencies.append(
                WorkItemDependency(
                    source_work_item_id=work_item_id,
                    target_work_item_id=target_id,
                    relationship_type=relationship,
                    relationship_type_raw=link_type,
                )
            )

    # Parse description for implicit references (best effort)
    description = _get(issue, "description") or ""
    if description:
        for match in _GITLAB_ISSUE_REF_PATTERN.finditer(str(description)):
            ref_project = match.group("project") or project_full_path
            ref_iid = match.group("iid")
            if not ref_iid:
                continue

            target_id = f"gitlab:{ref_project}#{ref_iid}"
            if target_id in seen_targets or target_id == work_item_id:
                continue
            seen_targets.add(target_id)

            # Check if reference is near blocking keywords
            start = max(0, match.start() - 50)
            context = description[start : match.end()].lower()
            relationship = "relates_to"
            relationship_raw = "description_reference"
            for kw in _BLOCKING_KEYWORDS:
                if kw in context:
                    relationship = "blocked_by" if "by" in kw else "blocks"
                    relationship_raw = kw
                    break

            dependencies.append(
                WorkItemDependency(
                    source_work_item_id=work_item_id,
                    target_work_item_id=target_id,
                    relationship_type=relationship,
                    relationship_type_raw=relationship_raw,
                )
            )

    return dependencies


def gitlab_milestone_to_sprint(
    *,
    milestone: Any,
    project_full_path: str,
) -> Sprint:
    """
    Convert a GitLab milestone to a Sprint model.

    GitLab milestones serve as sprint boundaries.
    """
    ms_id = _get(milestone, "id")
    title = _get(milestone, "title") or ""
    state = _get(milestone, "state") or "active"

    start_date = _to_utc(_parse_iso(_get(milestone, "start_date")))
    due_date = _to_utc(_parse_iso(_get(milestone, "due_date")))

    # Determine sprint state
    if state == "closed":
        sprint_state = "closed"
    elif state == "active":
        sprint_state = "active"
    else:
        sprint_state = "future"

    return Sprint(
        sprint_id=f"gitlab:{project_full_path}:milestone:{ms_id}",
        provider="gitlab",
        name=str(title),
        state=sprint_state,
        started_at=start_date,
        ended_at=due_date,
        completed_at=due_date if state == "closed" else None,
    )


def enrich_work_item_with_priority(
    work_item: WorkItem,
    labels: Sequence[str],
) -> WorkItem:
    """
    Enrich a WorkItem with priority_raw and service_class from labels.

    Returns a new WorkItem with updated fields.
    """
    if work_item.priority_raw is not None:
        return work_item

    priority_raw, service_class = _priority_from_labels(labels)
    if priority_raw is None:
        return work_item

    return WorkItem(
        work_item_id=work_item.work_item_id,
        provider=work_item.provider,
        repo_id=work_item.repo_id,
        project_key=work_item.project_key,
        project_id=work_item.project_id,
        title=work_item.title,
        description=work_item.description,
        type=work_item.type,
        status=work_item.status,
        status_raw=work_item.status_raw,
        assignees=work_item.assignees,
        reporter=work_item.reporter,
        created_at=work_item.created_at,
        updated_at=work_item.updated_at,
        started_at=work_item.started_at,
        completed_at=work_item.completed_at,
        closed_at=work_item.closed_at,
        labels=work_item.labels,
        story_points=work_item.story_points,
        sprint_id=work_item.sprint_id,
        sprint_name=work_item.sprint_name,
        url=work_item.url,
        priority_raw=priority_raw,
        service_class=service_class,
    )



from __future__ import annotations

import pytest

from dev_health_ops.llm.providers.openai import OpenAIProvider


class _StubResponse:
    def __init__(self) -> None:
        self.output_text = '{"status": "ok"}'


class _StubResponses:
    def __init__(self, captured: dict) -> None:
        self._captured = captured

    async def create(self, **kwargs):
        self._captured["kwargs"] = kwargs
        return _StubResponse()


class _StubClient:
    def __init__(self, captured: dict) -> None:
        self.responses = _StubResponses(captured)


@pytest.mark.asyncio
async def test_openai_provider_uses_max_completion_tokens_for_gpt5():
    captured: dict = {}
    provider = OpenAIProvider(
        api_key="test", model="gpt-5-mini", max_completion_tokens=123
    )
    provider._impl._client = _StubClient(captured)

    result = await provider.complete("hello")
    assert result == '{"status": "ok"}'
    # GPT-5 internal parameter is max_output_tokens
    assert "max_output_tokens" in captured["kwargs"]
    assert (
        captured["kwargs"]["max_output_tokens"] == 4096
    )  # max(clamped 2048 explanation default 4096)
    assert "max_tokens" not in captured["kwargs"]
    assert "temperature" not in captured["kwargs"]



"""GraphQL DataLoaders for batching and caching data fetches."""

from .base import CachedDataLoader, SimpleDataLoader, make_cache_key
from .analytics_loader import (
    DataLoaders,
    TimeseriesLoader,
    BreakdownLoader,
    TimeseriesKey,
    BreakdownKey,
    TimeseriesResultData,
    TimeseriesBucketData,
    BreakdownResultData,
    BreakdownItemData,
)
from .dimension_loader import (
    load_dimension_values,
    get_dimension_descriptions,
    get_measure_descriptions,
)
from .team_loader import TeamLoader, TeamByNameLoader, TeamData
from .repo_loader import RepoLoader, RepoByNameLoader, RepoData

__all__ = [
    # Base classes
    "CachedDataLoader",
    "SimpleDataLoader",
    "make_cache_key",
    # Analytics loaders
    "DataLoaders",
    "TimeseriesLoader",
    "BreakdownLoader",
    "TimeseriesKey",
    "BreakdownKey",
    "TimeseriesResultData",
    "TimeseriesBucketData",
    "BreakdownResultData",
    "BreakdownItemData",
    # Dimension loader
    "load_dimension_values",
    "get_dimension_descriptions",
    "get_measure_descriptions",
    # Entity loaders
    "TeamLoader",
    "TeamByNameLoader",
    "TeamData",
    "RepoLoader",
    "RepoByNameLoader",
    "RepoData",
]



"""Strict JSON schema validation for compute-time investment categorization."""

from __future__ import annotations

import json
import math
from dataclasses import dataclass
from typing import Dict, List, Optional

from dev_health_ops.work_graph.investment.taxonomy import SUBCATEGORIES
from dev_health_ops.work_graph.investment.utils import ensure_full_subcategory_vector

ALLOWED_TOP_LEVEL_KEYS = {"subcategories", "evidence_quotes", "uncertainty"}
ALLOWED_QUOTE_KEYS = {"quote", "source", "id"}
ALLOWED_SOURCES = {"issue", "pr", "commit"}

MIN_SUM = 0.98
MAX_SUM = 1.02
MAX_UNCERTAINTY_LEN = 280
MAX_QUOTE_LEN = 280
MIN_QUOTES = 1
MAX_QUOTES = 10


@dataclass(frozen=True)
class EvidenceQuote:
    quote: str
    source_type: str
    source_id: str


@dataclass(frozen=True)
class LLMValidationResult:
    ok: bool
    errors: List[str]
    subcategories: Dict[str, float]
    evidence_quotes: List[EvidenceQuote]
    uncertainty: str


def parse_llm_json(raw_text: str) -> tuple[Optional[Dict[str, object]], List[str]]:
    try:
        payload = json.loads(raw_text)
    except json.JSONDecodeError as exc:
        return None, [f"invalid_json: {exc}"]
    if not isinstance(payload, dict):
        return None, ["payload_not_object"]
    return payload, []


def validate_llm_payload(
    payload: Dict[str, object],
    source_texts: Dict[str, Dict[str, str]],
) -> LLMValidationResult:
    errors: List[str] = []

    keys = set(payload.keys())
    if keys != ALLOWED_TOP_LEVEL_KEYS:
        missing = ALLOWED_TOP_LEVEL_KEYS - keys
        extra = keys - ALLOWED_TOP_LEVEL_KEYS
        if missing:
            errors.append(f"missing_top_level_keys:{sorted(missing)}")
        if extra:
            errors.append(f"unexpected_top_level_keys:{sorted(extra)}")

    raw_subcategories = payload.get("subcategories")
    if not isinstance(raw_subcategories, dict):
        errors.append("subcategories_not_object")
        raw_subcategories = {}

    cleaned: Dict[str, float] = {}
    for key, value in raw_subcategories.items():
        if key not in SUBCATEGORIES:
            errors.append(f"unknown_subcategory:{key}")
            continue
        try:
            numeric = float(value)
        except (TypeError, ValueError):
            errors.append(f"invalid_probability:{key}")
            continue
        if not math.isfinite(numeric) or numeric < 0.0 or numeric > 1.0:
            errors.append(f"probability_out_of_range:{key}")
            continue
        cleaned[key] = numeric

    total = sum(cleaned.values())
    if total < MIN_SUM or total > MAX_SUM:
        errors.append(f"probability_sum_out_of_range:{total:.4f}")

    evidence_quotes_raw = payload.get("evidence_quotes")
    evidence_quotes: List[EvidenceQuote] = []
    if not isinstance(evidence_quotes_raw, list):
        errors.append("evidence_quotes_not_list")
        evidence_quotes_raw = []

    if isinstance(evidence_quotes_raw, list):
        if (
            len(evidence_quotes_raw) < MIN_QUOTES
            or len(evidence_quotes_raw) > MAX_QUOTES
        ):
            errors.append("evidence_quotes_count_out_of_range")
        for idx, entry in enumerate(evidence_quotes_raw):
            if not isinstance(entry, dict):
                errors.append(f"evidence_quote_not_object:{idx}")
                continue
            entry_keys = set(entry.keys())
            if entry_keys != ALLOWED_QUOTE_KEYS:
                missing = ALLOWED_QUOTE_KEYS - entry_keys
                extra = entry_keys - ALLOWED_QUOTE_KEYS
                if missing:
                    errors.append(
                        f"evidence_quote_missing_keys:{idx}:{sorted(missing)}"
                    )
                if extra:
                    errors.append(f"evidence_quote_extra_keys:{idx}:{sorted(extra)}")
                continue
            quote = str(entry.get("quote") or "").strip()
            source_type = str(entry.get("source") or "").strip()
            source_id = str(entry.get("id") or "").strip()
            if not quote:
                errors.append(f"evidence_quote_empty:{idx}")
                continue
            if len(quote) > MAX_QUOTE_LEN:
                errors.append(f"evidence_quote_too_long:{idx}")
                continue
            if source_type not in ALLOWED_SOURCES:
                errors.append(f"evidence_quote_invalid_source:{idx}:{source_type}")
                continue
            if not source_id:
                errors.append(f"evidence_quote_missing_id:{idx}")
                continue
            source_map = source_texts.get(source_type) or {}
            source_text = source_map.get(source_id, "")
            if not source_text:
                errors.append(f"evidence_quote_unknown_source:{idx}")
                continue
            if quote not in source_text:
                errors.append(f"evidence_quote_not_substring:{idx}")
                continue
            evidence_quotes.append(
                EvidenceQuote(
                    quote=quote,
                    source_type=source_type,
                    source_id=source_id,
                )
            )

    uncertainty_raw = payload.get("uncertainty")
    uncertainty = str(uncertainty_raw or "").strip()
    if not uncertainty:
        errors.append("uncertainty_missing")
    elif len(uncertainty) > MAX_UNCERTAINTY_LEN:
        errors.append("uncertainty_too_long")

    if errors:
        return LLMValidationResult(
            ok=False,
            errors=errors,
            subcategories={},
            evidence_quotes=[],
            uncertainty=uncertainty,
        )

    normalized = ensure_full_subcategory_vector(cleaned)
    return LLMValidationResult(
        ok=True,
        errors=[],
        subcategories=normalized,
        evidence_quotes=evidence_quotes,
        uncertainty=uncertainty,
    )



import hashlib
import logging
import os
import uuid
from datetime import datetime, timezone

from git import Repo as GitRepo
from sqlalchemy import (
    JSON,
    Boolean,
    Column,
    DateTime,
    ForeignKey,
    ForeignKeyConstraint,
    Integer,
    Text,
    TypeDecorator,
    CHAR,
)
from sqlalchemy.dialects.postgresql import UUID as PGUUID
from sqlalchemy.orm import declarative_base, relationship
from typing import Optional, List, Tuple

Base = declarative_base()


class GUID(TypeDecorator):
    """Platform-independent GUID type.

    Uses PostgreSQL's UUID type for PostgreSQL, otherwise uses
    CHAR(32), storing as stringified hex values.
    """

    impl = CHAR
    cache_ok = True

    def load_dialect_impl(self, dialect):
        if dialect.name == "postgresql":
            return dialect.type_descriptor(PGUUID())
        else:
            return dialect.type_descriptor(CHAR(32))

    def process_bind_param(self, value, dialect):
        if value is None:
            return value
        elif dialect.name == "postgresql":
            return str(value)
        else:
            if not isinstance(value, uuid.UUID):
                return str(uuid.UUID(value)).replace("-", "")
            else:
                # hex string
                return value.hex

    def process_result_value(self, value, dialect):
        if value is None:
            return value
        else:
            if not isinstance(value, uuid.UUID):
                return uuid.UUID(value)
            else:
                return value


def get_repo_uuid_from_repo(repo: str) -> uuid.UUID:
    """Generate a deterministic UUID from a repo identifier string.

    This is used for non-local repositories where we don't have a filesystem
    path to derive an ID from.

    Priority order:
    1. REPO_UUID environment variable (if set)
    2. SHA256(repo identifier)
    """
    env_uuid = os.getenv("REPO_UUID")
    if env_uuid:
        return uuid.UUID(env_uuid)

    if not repo:
        raise ValueError("repo identifier is required")

    normalized = repo.strip().lower()
    hash_obj = hashlib.sha256(normalized.encode("utf-8"))
    uuid_bytes = hash_obj.digest()[:16]
    return uuid.UUID(bytes=uuid_bytes)


def get_repo_uuid(repo_path: str) -> uuid.UUID:
    """
    Generate a deterministic UUID for a repository based on git data.

    Priority order:
    1. REPO_UUID environment variable (if set)
    2. Remote URL (if repository has a remote configured)
    3. Absolute path of the repository (fallback)

    :param repo_path: Path to the git repository.
    :return: UUID for the repository.
    """
    # First check if REPO_UUID is explicitly set
    env_uuid = os.getenv("REPO_UUID")
    if env_uuid:
        return uuid.UUID(env_uuid)

    try:
        # Try to get repository information from git
        git_repo = GitRepo(repo_path)

        # Try to get remote URL (most reliable identifier)
        if git_repo.remotes:
            remote_url = None
            # Try to get the origin remote first
            if "origin" in [r.name for r in git_repo.remotes]:
                origin = git_repo.remote("origin")
                remote_url = list(origin.urls)[0] if origin.urls else None
            else:
                # Use first available remote
                remote_url = (
                    list(git_repo.remotes[0].urls)[0]
                    if git_repo.remotes[0].urls
                    else None
                )

            if remote_url:
                # Create deterministic UUID from remote URL
                # Use SHA256 hash and convert to UUID format
                hash_obj = hashlib.sha256(remote_url.encode("utf-8"))
                # Take first 16 bytes of hash and create UUID
                uuid_bytes = hash_obj.digest()[:16]
                return uuid.UUID(bytes=uuid_bytes)

        # Fallback to absolute path if no remote
        abs_path = os.path.abspath(repo_path)
        hash_obj = hashlib.sha256(abs_path.encode("utf-8"))
        uuid_bytes = hash_obj.digest()[:16]
        return uuid.UUID(bytes=uuid_bytes)

    except Exception as e:
        # If anything fails, generate a random UUID
        logging.warning(
            "Could not derive UUID from git repository data: %s. "
            "Generating random UUID.",
            e,
        )
        return uuid.uuid4()


class Repo(Base, GitRepo):
    __tablename__ = "repos"

    def __init__(self, repo_path: Optional[str] = None, **kwargs):
        """
        Initialize the Repo class with the given repository path.

        Automatically derives a deterministic UUID from the repository
        if one is not explicitly provided.

        :param repo_path: Path to the git repository (optional).
        :param kwargs: Additional keyword arguments for SQLAlchemy ORM.
        """
        # Auto-derive UUID if not explicitly provided.
        # This is critical for MongoDB upserts (Repo.id may otherwise be None
        # until SQLAlchemy flush/commit, causing all upserts to collide).
        if "id" not in kwargs:
            if repo_path:
                kwargs["id"] = get_repo_uuid(repo_path)
            else:
                repo_identifier = kwargs.get("repo")
                if repo_identifier:
                    try:
                        kwargs["id"] = get_repo_uuid_from_repo(repo_identifier)
                    except Exception as e:
                        logging.warning(
                            "Could not derive UUID from repo identifier: %s. "
                            "Generating random UUID.",
                            e,
                        )
                        kwargs["id"] = uuid.uuid4()

        # Move 'tags' to 'repo_tags' to avoid conflict with GitRepo.tags property
        if "tags" in kwargs:
            kwargs["repo_tags"] = kwargs.pop("tags")

        super().__init__(**kwargs)  # Initialize SQLAlchemy ORM
        if repo_path:
            GitRepo.__init__(self, repo_path)  # Initialize GitPython Repo

    id = Column(
        GUID,
        primary_key=True,
        default=uuid.uuid4,
        comment="MergeStat identifier for the repo",
    )
    repo = Column(Text, nullable=False, comment="URL for the repo")
    ref = Column(Text, comment="ref for the repo")
    created_at = Column(
        DateTime(timezone=True),
        nullable=False,
        default=lambda: datetime.now(timezone.utc),
        comment="timestamp of when the MergeStat repo entry was created",
    )
    settings = Column(
        JSON, nullable=False, default=dict, comment="JSON settings for the repo"
    )
    repo_tags = Column(
        "tags",
        JSON,
        key="repo_tags",
        nullable=False,
        default=list,
        comment="array of tags for the repo",
    )
    # repo_import_id = Column(
    #     GUID,
    #     ForeignKey("mergestat.repo_imports.id", ondelete="CASCADE"),
    #     comment="foreign key for mergestat.repo_imports.id",
    # )
    # provider = Column(
    #     GUID,
    #     ForeignKey("mergestat.providers.id", ondelete="CASCADE"),
    #     nullable=False,
    # )

    # Relationships
    git_refs = relationship("GitRef", back_populates="repo")
    git_files = relationship("GitFile", back_populates="repo")
    git_commits = relationship("GitCommit", back_populates="repo")
    git_commit_stats = relationship("GitCommitStat", back_populates="repo")
    git_blames = relationship("GitBlame", back_populates="repo")
    git_pull_requests = relationship("GitPullRequest", back_populates="repo")
    ci_pipeline_runs = relationship(
        "CiPipelineRun",
        back_populates="repo",
        cascade="all, delete-orphan",
    )
    deployments = relationship(
        "Deployment",
        back_populates="repo",
        cascade="all, delete-orphan",
    )
    incidents = relationship(
        "Incident",
        back_populates="repo",
        cascade="all, delete-orphan",
    )


class GitRef(Base):
    __tablename__ = "git_refs"
    repo_id = Column(
        GUID,
        ForeignKey("repos.id", ondelete="CASCADE"),
        primary_key=True,
        comment="foreign key for public.repos.id",
    )
    full_name = Column(Text, primary_key=True)
    hash = Column(Text, comment="hash of the commit for refs that are not of type tag")
    name = Column(Text, comment="name of the ref")
    remote = Column(Text, comment="remote of the ref")
    target = Column(Text, comment="target of the ref")
    type = Column(Text, comment="type of the ref")
    tag_commit_hash = Column(
        Text, comment="hash of the commit for refs that are of type tag"
    )
    last_synced = Column(
        DateTime(timezone=True),
        nullable=False,
        default=lambda: datetime.now(timezone.utc),
        comment="timestamp when record was synced into the MergeStat database",
    )

    # Relationships
    repo = relationship("Repo", back_populates="git_refs")


class GitFile(Base):
    __tablename__ = "git_files"
    repo_id = Column(
        GUID,
        ForeignKey("repos.id", ondelete="CASCADE"),
        primary_key=True,
        comment="foreign key for public.repos.id",
    )
    path = Column(Text, primary_key=True, comment="path of the file")
    executable = Column(
        Boolean,
        nullable=False,
        comment="boolean to determine if the file is an executable",
    )
    contents = Column(Text, comment="contents of the file")
    last_synced = Column(
        DateTime(timezone=True),
        nullable=False,
        default=lambda: datetime.now(timezone.utc),
        comment="timestamp when record was synced into the MergeStat database",
    )

    # Relationships
    repo = relationship("Repo", back_populates="git_files")


class GitCommit(Base):
    __tablename__ = "git_commits"
    repo_id = Column(
        GUID,
        ForeignKey("repos.id", ondelete="CASCADE"),
        primary_key=True,
        comment="foreign key for public.repos.id",
    )
    hash = Column(Text, primary_key=True, comment="hash of the commit")
    message = Column(Text, comment="message of the commit")
    author_name = Column(Text, comment="name of the author of the modification")
    author_email = Column(Text, comment="email of the author of the modification")
    author_when = Column(
        DateTime(timezone=True),
        nullable=False,
        comment="timestamp of when the modification was authored",
    )
    committer_name = Column(
        Text, comment="name of the author who committed the modification"
    )
    committer_email = Column(
        Text, comment="email of the author who committed the modification"
    )
    committer_when = Column(
        DateTime(timezone=True),
        nullable=False,
        comment="timestamp of when the commit was made",
    )
    parents = Column(
        Integer, nullable=False, comment="the number of parents of the commit"
    )
    last_synced = Column(
        DateTime(timezone=True),
        nullable=False,
        default=lambda: datetime.now(timezone.utc),
        comment="timestamp when record was synced into the MergeStat database",
    )

    # Relationships
    repo = relationship("Repo", back_populates="git_commits")


class GitCommitStat(Base):
    __tablename__ = "git_commit_stats"
    repo_id = Column(
        GUID,
        ForeignKey("repos.id", ondelete="CASCADE"),
        primary_key=True,
        comment="foreign key for public.repos.id",
    )
    commit_hash = Column(Text, primary_key=True, comment="hash of the commit")
    file_path = Column(
        Text, primary_key=True, comment="path of the file the modification was made in"
    )
    additions = Column(
        Integer,
        nullable=False,
        comment="the number of additions in this path of the commit",
    )
    deletions = Column(
        Integer,
        nullable=False,
        comment="the number of deletions in this path of the commit",
    )
    old_file_mode = Column(
        Text,
        nullable=False,
        default="unknown",
        comment="old file mode derived from git mode",
    )
    new_file_mode = Column(
        Text, default="unknown", comment="new file mode derived from git mode"
    )
    last_synced = Column(
        DateTime(timezone=True),
        nullable=False,
        default=lambda: datetime.now(timezone.utc),
        comment="timestamp when record was synced into the MergeStat database",
    )

    # Relationships
    repo = relationship("Repo", back_populates="git_commit_stats")


class GitBlameMixin:
    """
    Mixin to provide functionality for fetching blame data using gitpython.
    """

    @staticmethod
    def fetch_blame(
        repo_path: str,
        filepath: str,
        repo_uuid: uuid.UUID,
        repo: Optional[GitRepo] = None,
    ) -> List[Tuple]:
        """
        Fetch blame data for a given file using gitpython.

        :param repo_path: Path to the git repository.
        :param filepath: Path to the file to fetch blame data for.
        :param repo_uuid: UUID of the repository.
        :param repo: Optional existing Repo instance to reuse (improves performance).
        :return: List of blame data tuples.
        """
        blame_data = []
        if repo is None:
            repo = GitRepo(repo_path)
        rel_path = os.path.relpath(filepath, repo_path)
        try:
            blame_info = repo.blame("HEAD", rel_path)
            if blame_info:
                line_no = 1
                for item in blame_info:
                    if not isinstance(item, (list, tuple)) or len(item) < 2:
                        continue
                    commit = item[0]
                    lines = item[1]
                    if commit is None or lines is None:
                        continue
                    for line in lines:
                        author_email = getattr(commit.author, "email", "unknown")
                        author_name = getattr(commit.author, "name", "unknown")
                        committed_datetime = getattr(
                            commit, "committed_datetime", datetime.now(timezone.utc)
                        )
                        hexsha = getattr(commit, "hexsha", "unknown")

                        blame_data.append(
                            (
                                repo_uuid,
                                author_email,
                                author_name,
                                committed_datetime,
                                hexsha,
                                line_no,
                                line.rstrip("\n") if line else "",
                                rel_path,
                            )
                        )
                        line_no += 1
        except Exception as e:
            logging.warning(f"Error processing {rel_path}: {e}")
        return blame_data


class GitBlame(Base, GitBlameMixin):
    __tablename__ = "git_blame"
    repo_id = Column(
        GUID,
        ForeignKey("repos.id", ondelete="CASCADE"),
        primary_key=True,
        comment="foreign key for public.repos.id",
    )
    path = Column(
        Text, primary_key=True, comment="path of the file the modification was made in"
    )
    line_no = Column(
        Integer, primary_key=True, comment="line number of the modification"
    )
    author_email = Column(Text, comment="email of the author who modified the line")
    author_name = Column(Text, comment="name of the author who modified the line")
    author_when = Column(
        DateTime(timezone=True),
        comment="timestamp of when the modification was authored",
    )
    commit_hash = Column(
        Text, comment="hash of the commit the modification was made in"
    )
    line = Column(Text, comment="content of the line")
    last_synced = Column(
        DateTime(timezone=True),
        nullable=False,
        default=lambda: datetime.now(timezone.utc),
        comment="timestamp when record was synced into the MergeStat database",
    )

    # Relationships
    repo = relationship("Repo", back_populates="git_blames")

    @classmethod
    def process_file(cls, repo_path, filepath, repo_uuid, repo=None):
        """
        Process a file to fetch blame data and return it as a list of GitBlame objects.

        :param repo_path: Path to the git repository.
        :param filepath: Path to the file to process.
        :param repo_uuid: UUID of the repository.
        :param repo: Optional existing Repo instance to reuse (improves performance).
        :return: List of GitBlame objects.
        """
        blame_data = cls.fetch_blame(repo_path, filepath, repo_uuid, repo=repo)
        return [
            cls(
                repo_id=row[0],
                author_email=row[1],
                author_name=row[2],
                author_when=row[3],
                commit_hash=row[4],
                line_no=row[5],
                line=row[6],
                path=row[7],
            )
            for row in blame_data
        ]


class GitPullRequest(Base):
    __tablename__ = "git_pull_requests"
    repo_id = Column(
        GUID,
        ForeignKey("repos.id", ondelete="CASCADE"),
        primary_key=True,
        comment="foreign key for public.repos.id",
    )
    number = Column(Integer, primary_key=True, comment="pull request number")
    title = Column(Text, comment="title of the pull request")
    body = Column(Text, comment="description/body of the pull request")
    state = Column(Text, comment="state of the pull request (open, closed, merged)")
    author_name = Column(Text, comment="username of the author")
    author_email = Column(Text, comment="email of the author (if available)")
    created_at = Column(
        DateTime(timezone=True),
        nullable=False,
        comment="timestamp when PR was created",
    )
    merged_at = Column(
        DateTime(timezone=True),
        comment="timestamp when PR was merged",
    )
    closed_at = Column(
        DateTime(timezone=True),
        comment="timestamp when PR was closed",
    )
    head_branch = Column(Text, comment="name of the head branch")
    base_branch = Column(Text, comment="name of the base branch")
    additions = Column(Integer, comment="total line additions in the PR")
    deletions = Column(Integer, comment="total line deletions in the PR")
    changed_files = Column(Integer, comment="total number of files changed in the PR")
    first_review_at = Column(
        DateTime(timezone=True), comment="timestamp of the first review"
    )
    first_comment_at = Column(
        DateTime(timezone=True), comment="timestamp of the first comment"
    )
    changes_requested_count = Column(
        Integer, default=0, comment="number of times changes were requested"
    )
    reviews_count = Column(Integer, default=0, comment="total number of reviews")
    comments_count = Column(Integer, default=0, comment="total number of comments")
    last_synced = Column(
        DateTime(timezone=True),
        nullable=False,
        default=lambda: datetime.now(timezone.utc),
        comment="timestamp when record was synced into the MergeStat database",
    )

    # Relationships
    repo = relationship("Repo", back_populates="git_pull_requests")
    reviews = relationship(
        "GitPullRequestReview",
        primaryjoin="and_(GitPullRequest.repo_id==GitPullRequestReview.repo_id, GitPullRequest.number==GitPullRequestReview.number)",
        foreign_keys="[GitPullRequestReview.repo_id, GitPullRequestReview.number]",
        back_populates="pr",
        cascade="all, delete-orphan",
    )


class GitPullRequestReview(Base):
    __tablename__ = "git_pull_request_reviews"
    repo_id = Column(
        GUID,
        ForeignKey("repos.id", ondelete="CASCADE"),
        primary_key=True,
    )
    number = Column(
        Integer,
        primary_key=True,
    )
    review_id = Column(
        Text,
        primary_key=True,
        comment="unique identifier for the review (e.g. GitHub review ID)",
    )
    reviewer = Column(Text, nullable=False, comment="identity of the reviewer")
    state = Column(
        Text,
        nullable=False,
        comment="state of the review (APPROVED, CHANGES_REQUESTED, etc.)",
    )
    submitted_at = Column(
        DateTime(timezone=True),
        nullable=False,
        comment="timestamp when review was submitted",
    )
    last_synced = Column(
        DateTime(timezone=True),
        nullable=False,
        default=lambda: datetime.now(timezone.utc),
    )

    # Relationships
    pr = relationship(
        "GitPullRequest",
        primaryjoin="and_(GitPullRequestReview.repo_id==GitPullRequest.repo_id, GitPullRequestReview.number==GitPullRequest.number)",
        foreign_keys="[GitPullRequestReview.repo_id, GitPullRequestReview.number]",
        back_populates="reviews",
    )

    __table_args__ = (
        ForeignKeyConstraint(
            ["repo_id", "number"],
            ["git_pull_requests.repo_id", "git_pull_requests.number"],
            ondelete="CASCADE",
        ),
    )


class CiPipelineRun(Base):
    __tablename__ = "ci_pipeline_runs"
    repo_id = Column(
        GUID,
        ForeignKey("repos.id", ondelete="CASCADE"),
        primary_key=True,
    )
    run_id = Column(Text, primary_key=True)
    status = Column(Text)
    queued_at = Column(DateTime(timezone=True))
    started_at = Column(DateTime(timezone=True), nullable=False)
    finished_at = Column(DateTime(timezone=True))
    last_synced = Column(
        DateTime(timezone=True),
        nullable=False,
        default=lambda: datetime.now(timezone.utc),
    )

    repo = relationship("Repo", back_populates="ci_pipeline_runs")


class Deployment(Base):
    __tablename__ = "deployments"
    repo_id = Column(
        GUID,
        ForeignKey("repos.id", ondelete="CASCADE"),
        primary_key=True,
    )
    deployment_id = Column(Text, primary_key=True)
    status = Column(Text)
    environment = Column(Text)
    started_at = Column(DateTime(timezone=True))
    finished_at = Column(DateTime(timezone=True))
    deployed_at = Column(DateTime(timezone=True))
    merged_at = Column(DateTime(timezone=True))
    pull_request_number = Column(Integer)
    last_synced = Column(
        DateTime(timezone=True),
        nullable=False,
        default=lambda: datetime.now(timezone.utc),
    )

    repo = relationship("Repo", back_populates="deployments")


class Incident(Base):
    __tablename__ = "incidents"
    repo_id = Column(
        GUID,
        ForeignKey("repos.id", ondelete="CASCADE"),
        primary_key=True,
    )
    incident_id = Column(Text, primary_key=True)
    status = Column(Text)
    started_at = Column(DateTime(timezone=True), nullable=False)
    resolved_at = Column(DateTime(timezone=True))
    last_synced = Column(
        DateTime(timezone=True),
        nullable=False,
        default=lambda: datetime.now(timezone.utc),
    )

    repo = relationship("Repo", back_populates="incidents")
